import dataclasses
import uuid

from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.gui.preset_settings.item_pool_tab import PresetItemPool
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.major_item_state import MajorItemState


def test_on_default_item_updated(skip_qtbot, preset_manager):
    # Setup
    game = RandovaniaGame.METROID_PRIME_ECHOES
    base = preset_manager.default_preset_for_game(game).get_preset()
    preset = dataclasses.replace(base,
                                 uuid=uuid.UUID('b41fde84-1f57-4b79-8cd6-3e5a78077fa6'),
                                 base_preset_uuid=base.uuid)

    editor = PresetEditor(preset)
    window = PresetItemPool(editor)

    item_database = default_database.item_database_for_game(window.game)
    item = item_database.major_items["Dark Beam"]
    category = item_database.item_categories["beam"]
    combo = window._default_items[category]

    # Run
    combo.setCurrentIndex(combo.findData(item))

    # Assert
    assert editor.major_items_configuration.default_items[category] == item
    assert editor.major_items_configuration.items_state[item] == MajorItemState(
        num_included_in_starting_items=1,
        included_ammo=(50,)
    )



import dataclasses
import json
import uuid
from typing import Optional, List, Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder, BitPackValue
from randovania.games.game import RandovaniaGame
from randovania.layout.game_to_class import AnyGameConfiguration, GAME_TO_CONFIGURATION


def _dictionary_byte_hash(data: dict) -> int:
    return bitpacking.single_byte_hash(json.dumps(data, separators=(',', ':')).encode("UTF-8"))


@dataclasses.dataclass(frozen=True)
class Preset(BitPackValue):
    name: str
    uuid: uuid.UUID
    description: str
    base_preset_uuid: Optional[uuid.UUID]
    game: RandovaniaGame
    configuration: AnyGameConfiguration

    @property
    def as_json(self) -> dict:
        return {
            "name": self.name,
            "uuid": str(self.uuid),
            "description": self.description,
            "base_preset_uuid": str(self.base_preset_uuid) if self.base_preset_uuid is not None else None,
            "game": self.game.value,
            "configuration": self.configuration.as_json,
        }

    @classmethod
    def from_json_dict(cls, value) -> "Preset":
        game = RandovaniaGame(value["game"])
        return Preset(
            name=value["name"],
            uuid=uuid.UUID(value["uuid"]),
            description=value["description"],
            base_preset_uuid=uuid.UUID(value["base_preset_uuid"]) if value["base_preset_uuid"] is not None else None,
            game=game,
            configuration=GAME_TO_CONFIGURATION[game].from_json(value["configuration"]),
        )

    def dangerous_settings(self) -> List[str]:
        return self.configuration.dangerous_settings()

    def is_same_configuration(self, other: "Preset") -> bool:
        return self.configuration == other.configuration

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        from randovania.interface_common.preset_manager import PresetManager
        manager: PresetManager = metadata["manager"]

        # Is this a custom preset?
        is_custom_preset = self.base_preset_uuid is not None
        reference = self
        if is_custom_preset:
            while reference.base_preset_uuid is not None:
                reference_versioned = manager.preset_for_uuid(reference.base_preset_uuid)
                if reference_versioned is None:
                    reference_versioned = manager.default_preset_for_game(reference.game)
                reference = reference_versioned.get_preset()

        included_preset_uuids = [versioned.uuid for versioned in manager.included_presets.values()]
        yield from bitpacking.encode_bool(is_custom_preset)
        yield from bitpacking.pack_array_element(reference.uuid, included_preset_uuids)
        if is_custom_preset:
            yield from self.configuration.bit_pack_encode({"reference": reference.configuration})
        yield _dictionary_byte_hash(self.configuration.game_data), 256

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "Preset":
        from randovania.interface_common.preset_manager import PresetManager
        manager: PresetManager = metadata["manager"]

        included_presets = [versioned for versioned in manager.included_presets.values()]

        is_custom_preset = bitpacking.decode_bool(decoder)
        reference = decoder.decode_element(included_presets).get_preset()
        if is_custom_preset:
            preset = Preset(
                name="{} Custom".format(reference.name),
                description="A customized preset.",
                uuid=uuid.uuid4(),
                base_preset_uuid=reference.uuid,
                game=reference.game,
                configuration=reference.configuration.bit_pack_unpack(decoder, {"reference": reference.configuration}),
            )
        else:
            preset = reference

        included_data_hash = decoder.decode_single(256)
        expected_data_hash = _dictionary_byte_hash(preset.configuration.game_data)
        if included_data_hash != expected_data_hash:
            raise ValueError("Given permalink is for a Randovania database with hash '{}', "
                             "but current database has hash '{}'.".format(included_data_hash, expected_data_hash))
        return preset

    def fork(self) -> "Preset":
        return dataclasses.replace(self, name=f"{self.name} Copy",
                                   description=f"A copy version of {self.name}.",
                                   uuid=uuid.uuid4(), base_preset_uuid=self.uuid)



import typing

from randovania.game_description import node_search
from randovania.game_description.world.area import Area
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint, HintLocationPrecision, RelativeDataArea, HintRelativeAreaName
from randovania.game_description.world.node import PickupNode
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.world_list import WorldList
from randovania.games.prime.patcher_file_lib import hint_lib


class LocationFormatter:
    def format(self, determiner: hint_lib.Determiner, pickup_name: str, hint: Hint) -> str:
        raise NotImplementedError()


class GuardianFormatter(LocationFormatter):
    _GUARDIAN_NAMES = {
        PickupIndex(43): "Amorbis",
        PickupIndex(79): "Chykka",
        PickupIndex(115): "Quadraxis",
    }

    def format(self, determiner: hint_lib.Determiner, pickup: str, hint: Hint) -> str:
        guardian = hint_lib.color_text(hint_lib.TextColor.GUARDIAN, self._GUARDIAN_NAMES[hint.target])
        return f"{guardian} is guarding {determiner}{pickup}."


class TemplatedFormatter(LocationFormatter):
    def __init__(self, template: str, area_namer: hint_lib.AreaNamer):
        self.template = template
        self.hint_name_creator = area_namer

    def format(self, determiner: hint_lib.Determiner, pickup: str, hint: Hint) -> str:
        node_name = self.hint_name_creator.location_name(
            hint.target,
            hint.precision.location == HintLocationPrecision.WORLD_ONLY
        )
        return self.template.format(determiner=determiner,
                                    pickup=pickup,
                                    node=node_name)


class RelativeFormatter(LocationFormatter):
    def __init__(self, world_list: WorldList, patches: GamePatches):
        self.world_list = world_list
        self.patches = patches
        self._index_to_node = {
            node.pickup_index: node
            for node in world_list.all_nodes
            if isinstance(node, PickupNode)
        }

    def _calculate_distance(self, source_location: PickupIndex, target: Area) -> int:
        source = self._index_to_node[source_location]
        return node_search.distances_to_node(self.world_list, source,
                                             patches=self.patches, ignore_elevators=False)[target]

    def relative_format(self, determiner: hint_lib.Determiner, pickup: str, hint: Hint, other_area: Area, other_name: str,
                        ) -> str:
        distance = self._calculate_distance(hint.target, other_area) + (hint.precision.relative.distance_offset or 0)
        if distance == 1:
            distance_msg = "one room"
        else:
            precise_msg = "exactly " if hint.precision.relative.distance_offset is None else "up to "
            distance_msg = f"{precise_msg}{distance} rooms"

        return (f"{determiner.title}{pickup} can be found "
                f"{hint_lib.color_text(hint_lib.TextColor.LOCATION, distance_msg)} away from {other_name}.")

    def format(self, determiner: hint_lib.Determiner, pickup_name: str, hint: Hint) -> str:
        raise NotImplementedError()


class RelativeAreaFormatter(RelativeFormatter):
    def format(self, determiner: hint_lib.Determiner, pickup: str, hint: Hint) -> str:
        relative = typing.cast(RelativeDataArea, hint.precision.relative)
        other_area = self.world_list.area_by_area_location(relative.area_location)

        if relative.precision == HintRelativeAreaName.NAME:
            other_name = self.world_list.area_name(other_area)
        elif relative.precision == HintRelativeAreaName.FEATURE:
            raise NotImplementedError("HintRelativeAreaName.FEATURE not implemented")
        else:
            raise ValueError(f"Unknown precision: {relative.precision}")

        return self.relative_format(determiner, pickup, hint, other_area, other_name)



from randovania.games.prime.prime1_dol_patches import Prime1DolVersion

ALL_VERSIONS = [
    Prime1DolVersion(
        version="0-00",
        description="GC NTSC 0-00",
        build_string_address=0x803cc588,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.088 10/29/2002 2:21:25",
        sda2_base=0x805b1d20,
        sda13_base=0x805aebc0,
        cplayer_vtable=0x803d96e8,
        message_receiver_string_ref=0x803efb90,
    ),
    Prime1DolVersion(
        version="0-02",
        description="GC NTSC 0-02",
        build_string_address=0x803cd648,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.111 3/10/2003 17:56:21",
        sda2_base=0x805b2de0,
        sda13_base=0x805afc40,
        cplayer_vtable=0x803da7a8,
        message_receiver_string_ref=0x803f0ba8,
    ),
    Prime1DolVersion(
        version="pal",
        description="GC PAL",
        build_string_address=0x803b6924,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.110 2/4/2003 22:16:07",
        sda2_base=0x80473e60,
        sda13_base=0x80470c60,
        cplayer_vtable=0x803c4b88,
        message_receiver_string_ref=0x803d7a28,
    ),
]



import dataclasses

from randovania.games.game import RandovaniaGame
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches


@dataclasses.dataclass(frozen=True)
class PrimeCosmeticPatches(BaseCosmeticPatches):
    qol_cosmetic: bool = True
    open_map: bool = True

    @classmethod
    def default(cls) -> "PrimeCosmeticPatches":
        return cls()

    @classmethod
    def game(cls):
        return RandovaniaGame.METROID_PRIME



from random import Random
from typing import Tuple
from unittest.mock import MagicMock

import pytest

from randovania.game_description.item.ammo import AMMO_ITEM_CATEGORY
from randovania.game_description.item.item_category import USELESS_ITEM_CATEGORY
from randovania.game_description import default_database
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry, ResourceLock, PickupModel, \
    ConditionalResources, ResourceConversion
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.node import PickupNode
from randovania.games.game import RandovaniaGame
from randovania.games.prime.patcher_file_lib import pickup_exporter
from randovania.generator.item_pool import pickup_creator
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.base.major_item_state import MajorItemState
from randovania.layout.base.pickup_model import PickupModelStyle, PickupModelDataSource


def test_get_single_hud_text_all_major_items(echoes_item_database, echoes_resource_database):
    memo_data = default_database.default_prime2_memo_data()

    # Run
    for item in echoes_item_database.major_items.values():
        pickup = pickup_creator.create_major_item(item, MajorItemState(), False, echoes_resource_database, None, False)

        result = pickup_exporter._get_all_hud_text(pickup_exporter._conditional_resources_for_pickup(pickup),
                                                   memo_data)
        for i, progression in enumerate(pickup.progression):
            assert progression[0].long_name in result[i]
        assert result
        for line in result:
            assert len(line) > 10
            assert isinstance(line, str)


@pytest.mark.parametrize("order", [
    ("X", "Y"),
    ("Y", "X"),
    ("Y", "Z"),
])
def test_calculate_hud_text(order: Tuple[str, str], generic_item_category):
    # Setup
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    pickup_x = PickupEntry("A", 1, generic_item_category, generic_item_category,
                           progression=(
                               ((resource_a, 1),)
                           ))
    pickup_y = PickupEntry("Y", 2, generic_item_category, generic_item_category,
                           progression=(
                               (resource_b, 1),
                               (resource_a, 5),
                           ))
    pickup_z = PickupEntry("Z", 2, generic_item_category, generic_item_category,
                           progression=(
                               (resource_a, 1),
                               (resource_b, 5),
                           ))

    memo_data = {
        "A": "You got {A} of A",
        "B": "You found {B} of B",
    }
    pickups = {
        "X": pickup_x,
        "Y": pickup_y,
        "Z": pickup_z,
    }

    # Run
    result = pickup_exporter._calculate_hud_text(pickups[order[0]], pickups[order[1]],
                                                 PickupModelStyle.HIDE_ALL, memo_data)

    # Assert
    if order[1] == "Y":
        assert result == ["You found 1 of B"]
    elif order[1] == "X":
        assert result == ["You got 1 of A", "You got 1 of A"]
    else:
        assert result == ["You got 1 of A", "You found 5 of B"]


@pytest.mark.parametrize("model_style", PickupModelStyle)
def test_create_pickup_list(model_style: PickupModelStyle, empty_patches, generic_item_category):
    # Setup
    has_scan_text = model_style in {PickupModelStyle.ALL_VISIBLE, PickupModelStyle.HIDE_MODEL}
    rng = Random(5000)

    model_0 = MagicMock(spec=PickupModel)
    model_1 = MagicMock(spec=PickupModel)
    model_2 = MagicMock(spec=PickupModel)
    useless_model = PickupModel(
        game=RandovaniaGame.METROID_PRIME_ECHOES,
        name="EnergyTransferModule",
    )

    useless_resource = ItemResourceInfo(0, "Useless", "Useless", 10, None)
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)
    pickup_a = PickupEntry("P-A", model_1, generic_item_category, generic_item_category,
                           progression=((resource_a, 1),),
                           )
    pickup_b = PickupEntry("P-B", model_2, generic_item_category, generic_item_category,
                           progression=((resource_b, 1),
                                        (resource_a, 5)), )
    pickup_c = PickupEntry("P-C", model_2, AMMO_ITEM_CATEGORY, generic_item_category,
                           progression=tuple(),
                           extra_resources=((resource_b, 2), (resource_a, 1)),
                           unlocks_resource=True,
                           resource_lock=ResourceLock(resource_a, resource_a, useless_resource))

    useless_pickup = PickupEntry("P-Useless", model_0, USELESS_ITEM_CATEGORY, USELESS_ITEM_CATEGORY, progression=((useless_resource, 1),))
    patches = empty_patches.assign_pickup_assignment({
        PickupIndex(0): PickupTarget(pickup_a, 0),
        PickupIndex(2): PickupTarget(pickup_b, 0),
        PickupIndex(3): PickupTarget(pickup_a, 0),
        PickupIndex(4): PickupTarget(pickup_c, 0),
    })
    creator = pickup_exporter.PickupExporterSolo(pickup_exporter.GenericAcquiredMemo())

    world_list = MagicMock()
    world_list.all_nodes = [
        PickupNode(f"Name {i}", False, None, i, PickupIndex(i), False)
        for i in range(5)
    ]

    # Run
    result = pickup_exporter.export_all_indices(
        patches,
        PickupTarget(useless_pickup, 0),
        world_list,
        rng,
        model_style,
        PickupModelDataSource.ETM,
        creator,
        pickup_creator.create_visual_etm(),
    )

    # Assert
    assert len(result) == 5
    assert result[0] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(0),
        scan_text="P-A" if has_scan_text else "Unknown item",
        hud_text=["A acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        conditional_resources=[ConditionalResources("A", None, ((resource_a, 1),))],
        conversion=[],
        model=model_1 if model_style == PickupModelStyle.ALL_VISIBLE else useless_model,
        other_player=False,
    )
    assert result[1] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(1),
        scan_text="P-Useless" if has_scan_text else "Unknown item",
        hud_text=["Useless acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        conditional_resources=[ConditionalResources("Useless", None, ((useless_resource, 1),))],
        conversion=[],
        model=model_0 if model_style == PickupModelStyle.ALL_VISIBLE else useless_model,
        other_player=False,
    )
    assert result[2] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(2),
        scan_text="P-B. Provides the following in order: B, A" if has_scan_text else "Unknown item",
        hud_text=["B acquired!", "A acquired!"] if model_style != PickupModelStyle.HIDE_ALL else [
            'Unknown item acquired!', 'Unknown item acquired!'],
        conditional_resources=[
            ConditionalResources("B", None, ((resource_b, 1),)),
            ConditionalResources("A", resource_b, ((resource_a, 5),)),
        ],
        conversion=[],
        model=model_2 if model_style == PickupModelStyle.ALL_VISIBLE else useless_model,
        other_player=False,
    )
    assert result[3] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(3),
        scan_text="P-A" if has_scan_text else "Unknown item",
        hud_text=["A acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        conditional_resources=[ConditionalResources("A", None, ((resource_a, 1),))],
        conversion=[],
        model=model_1 if model_style == PickupModelStyle.ALL_VISIBLE else useless_model,
        other_player=False,
    )
    assert result[4] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(4),
        scan_text="P-C. Provides 2 B and 1 A" if has_scan_text else "Unknown item",
        hud_text=["P-C acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        conditional_resources=[ConditionalResources("P-C", None, (
            (resource_b, 2), (resource_a, 1),
        ))],
        conversion=[ResourceConversion(source=useless_resource, target=resource_a)],
        model=model_2 if model_style == PickupModelStyle.ALL_VISIBLE else useless_model,
        other_player=False,
    )


@pytest.mark.parametrize("has_memo_data", [False, True])
def test_create_pickup_list_random_data_source(has_memo_data: bool, empty_patches, generic_item_category):
    # Setup
    rng = Random(5000)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    model_1 = MagicMock(spec=PickupModel)
    model_2 = MagicMock(spec=PickupModel)
    useless_model = PickupModel(game=RandovaniaGame.METROID_PRIME_CORRUPTION, name="Useless")

    pickup_a = PickupEntry("A", model_1, generic_item_category, generic_item_category, 
                            progression=tuple())
    pickup_b = PickupEntry("B", model_2, generic_item_category, generic_item_category,
                            progression=((resource_b, 1), (resource_b, 1)))
    pickup_c = PickupEntry("C", model_2, generic_item_category, generic_item_category, 
                            progression=tuple())
    useless_pickup = PickupEntry("Useless", useless_model, USELESS_ITEM_CATEGORY, USELESS_ITEM_CATEGORY, progression=tuple())

    patches = empty_patches.assign_pickup_assignment({
        PickupIndex(0): PickupTarget(pickup_a, 0),
        PickupIndex(2): PickupTarget(pickup_b, 0),
        PickupIndex(3): PickupTarget(pickup_a, 0),
        PickupIndex(4): PickupTarget(pickup_c, 0),
    })

    if has_memo_data:
        memo_data = {
            "A": "This is an awesome item A",
            "B": "This is B. It is good.",
            "C": "What a nice day to have a C",
            "Useless": "Try again next time",
        }
    else:
        memo_data = {
            name: "{} acquired!".format(name)
            for name in ("A", "B", "C", "Useless")
        }

    creator = pickup_exporter.PickupExporterSolo(memo_data)

    world_list = MagicMock()
    world_list.all_nodes = [
        PickupNode(f"Name {i}", False, None, i, PickupIndex(i), False)
        for i in range(5)
    ]

    # Run
    result = pickup_exporter.export_all_indices(
        patches,
        PickupTarget(useless_pickup, 0),
        world_list,
        rng,
        PickupModelStyle.HIDE_ALL,
        PickupModelDataSource.RANDOM,
        creator,
        pickup_creator.create_visual_etm(),
    )

    # Assert
    assert len(result) == 5
    assert result[0] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(0),
        scan_text="A",
        hud_text=[memo_data["A"]],
        conditional_resources=[ConditionalResources("A", None, ())],
        conversion=[],
        model=model_1,
        other_player=False,
    )
    assert result[1] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(1),
        scan_text="A",
        hud_text=[memo_data["A"]],
        conditional_resources=[ConditionalResources("Useless", None, ())],
        conversion=[],
        model=model_1,
        other_player=False,
    )
    assert result[2] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(2),
        scan_text="C",
        hud_text=[memo_data["C"], memo_data["C"]],
        conditional_resources=[
            ConditionalResources("B", None, ((resource_b, 1),)),
            ConditionalResources("B", resource_b, ((resource_b, 1),)),
        ],
        conversion=[],
        model=model_2,
        other_player=False,
    )
    assert result[3] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(3),
        scan_text="B",
        hud_text=[memo_data["B"]],
        conditional_resources=[ConditionalResources("A", None, ())],
        conversion=[],
        model=model_2,
        other_player=False,
    )
    assert result[4] == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(4),
        scan_text="A",
        hud_text=[memo_data["A"]],
        conditional_resources=[ConditionalResources("C", None, ())],
        conversion=[],
        model=model_1,
        other_player=False,
    )


def test_pickup_scan_for_progressive_suit(echoes_item_database, echoes_resource_database):
    # Setup
    progressive_suit = echoes_item_database.major_items["Progressive Suit"]
    pickup = pickup_creator.create_major_item(progressive_suit, MajorItemState(), False, echoes_resource_database,
                                              None, False)

    # Run
    result = pickup_exporter._pickup_scan(pickup)

    # Assert
    assert result == "Progressive Suit. Provides the following in order: Dark Suit, Light Suit"


@pytest.mark.parametrize(["item", "ammo", "result"], [
    ("Beam Ammo Expansion", [4, 20], "Beam Ammo Expansion. Provides 4 Dark Ammo, 20 Light Ammo and 1 Item Percentage"),
    ("Missile Expansion", [4], "Missile Expansion. Provides 4 Missiles and 1 Item Percentage"),
])
def test_pickup_scan_for_ammo_expansion(echoes_item_database, echoes_resource_database, item, ammo, result):
    # Setup
    expansion = echoes_item_database.ammo[item]
    pickup = pickup_creator.create_ammo_expansion(expansion, ammo, False, echoes_resource_database)

    # Run
    result = pickup_exporter._pickup_scan(pickup)

    # Assert
    assert result == result


@pytest.fixture(name="pickup_for_create_pickup_data")
def _create_pickup_data(generic_item_category):
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)
    return PickupEntry("Cake", 1, generic_item_category, generic_item_category,
                       progression=(
                           (resource_a, 1),
                           (resource_b, 1),
                       ))


def test_solo_create_pickup_data(pickup_for_create_pickup_data):
    # Setup
    creator = pickup_exporter.PickupExporterSolo(pickup_exporter.GenericAcquiredMemo())
    model = MagicMock()
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    # Run
    data = creator.create_details(PickupIndex(10), PickupTarget(pickup_for_create_pickup_data, 0),
                                  pickup_for_create_pickup_data, PickupModelStyle.ALL_VISIBLE,
                                  "Scan Text", model)

    # Assert
    assert data == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(10),
        scan_text="Scan Text",
        hud_text=['A acquired!', 'B acquired!'],
        conditional_resources=[
            ConditionalResources("A", None, ((resource_a, 1),)),
            ConditionalResources("B", resource_a, ((resource_b, 1),)),
        ],
        conversion=[],
        model=model,
        other_player=False,
    )


def test_multi_create_pickup_data_for_self(pickup_for_create_pickup_data):
    # Setup
    solo = pickup_exporter.PickupExporterSolo(pickup_exporter.GenericAcquiredMemo())
    creator = pickup_exporter.PickupExporterMulti(solo, MagicMock(), PlayersConfiguration(0, {0: "You", 1: "Someone"}))
    model = MagicMock()
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    # Run
    data = creator.create_details(PickupIndex(10), PickupTarget(pickup_for_create_pickup_data, 0),
                                  pickup_for_create_pickup_data, PickupModelStyle.ALL_VISIBLE,
                                  "Scan Text", model)

    # Assert
    assert data == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(10),
        scan_text="Your Scan Text",
        hud_text=['A acquired!', 'B acquired!'],
        conditional_resources=[
            ConditionalResources("A", None, ((resource_a, 1),)),
            ConditionalResources("B", resource_a, ((resource_b, 1),)),
        ],
        conversion=[],
        model=model,
        other_player=False,
    )


def test_multi_create_pickup_data_for_other(pickup_for_create_pickup_data):
    # Setup
    multi = ItemResourceInfo(10, "Multiworld", "Multiworld", 30, None)
    solo = pickup_exporter.PickupExporterSolo(pickup_exporter.GenericAcquiredMemo())
    creator = pickup_exporter.PickupExporterMulti(solo, multi, PlayersConfiguration(0, {0: "You", 1: "Someone"}))
    model = MagicMock()
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    # Run
    data = creator.create_details(PickupIndex(10), PickupTarget(pickup_for_create_pickup_data, 1),
                                  pickup_for_create_pickup_data, PickupModelStyle.ALL_VISIBLE,
                                  "Scan Text", model)

    # Assert
    assert data == pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(10),
        scan_text="Someone's Scan Text",
        hud_text=['Sent Cake to Someone!'],
        conditional_resources=[
            ConditionalResources(None, None, ((multi, 11),)),
        ],
        conversion=[],
        model=model,
        other_player=True,
    )



from randovania.games.patchers import claris_patcher_file


def test_create_starting_popup_empty(default_echoes_preset, echoes_resource_database):
    starting_items = {}

    # Run
    result = claris_patcher_file._create_starting_popup(default_echoes_preset.configuration,
                                                        echoes_resource_database,
                                                        starting_items)

    # Assert
    assert result == []


def test_create_starting_popup_items(default_echoes_preset, echoes_resource_database):
    starting_items = {
        echoes_resource_database.get_item_by_name("Missile"): 15,
        echoes_resource_database.energy_tank: 3,
        echoes_resource_database.get_item_by_name("Dark Beam"): 1,
        echoes_resource_database.get_item_by_name("Screw Attack"): 1,
    }

    # Run
    result = claris_patcher_file._create_starting_popup(default_echoes_preset.configuration,
                                                        echoes_resource_database,
                                                        starting_items)

    # Assert
    assert result == [
        'Extra starting items:',
        'Dark Beam, 3 Energy Tank, 15 Missiles, Screw Attack'
    ]


def test_adjust_model_name(randomizer_data):
    # Setup
    patcher_data = {
        "pickups": [
            {"model_name": "DarkVisor"},
            {"model_name": "SkyTempleKey"},
            {"model_name": "MissileExpansion"},
            {"model_name": "prime1_Boost Ball"},
            {"model_name": "prime1_Plasma Beam"},
        ]
    }

    # Run
    claris_patcher_file.adjust_model_name(patcher_data, randomizer_data)

    # Assert
    assert patcher_data == {
        "pickups": [
            {"model_index": 11, "sound_index": 0, "jingle_index": 1},
            {"model_index": 38, "sound_index": 1, "jingle_index": 2},
            {"model_index": 22, "sound_index": 0, "jingle_index": 0},
            {"model_index": 18, "sound_index": 0, "jingle_index": 1},
            {"model_index": 30, "sound_index": 0, "jingle_index": 1},
        ]
    }



from typing import NamedTuple, Optional

from randovania.game_description.resources.item_resource_info import ItemResourceInfo


class DamageReduction(NamedTuple):
    inventory_item: Optional[ItemResourceInfo]
    damage_multiplier: float



import logging
import random
import struct
from typing import List, Optional, Dict

from PySide2 import QtWidgets, QtCore
from PySide2.QtWidgets import QMainWindow
from qasync import asyncSlot

from randovania.game_connection.connector.echoes_remote_connector import EchoesRemoteConnector
from randovania.game_connection.executor.memory_operation import MemoryOperation, MemoryOperationExecutor
from randovania.game_description import default_database
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.world.node import PickupNode
from randovania.games.game import RandovaniaGame
from randovania.games.prime import echoes_dol_versions, all_prime_dol_patches
from randovania.gui.generated.debug_backend_window_ui import Ui_DebugBackendWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.qt_network_client import handle_network_errors
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches
from randovania.network_common.admin_actions import SessionAdminUserAction


class DebugGameBackendChoice:
    @property
    def pretty_text(self):
        return "Debug"


def _echoes_powerup_address(item_index: int) -> int:
    powerups_offset = 0x58
    vector_data_offset = 0x4
    powerup_size = 0xc
    return 0xA00000 + (powerups_offset + vector_data_offset) + (item_index * powerup_size)


class DebugExecutorWindow(MemoryOperationExecutor, Ui_DebugBackendWindow):
    _connected: bool = False
    _timer: QtCore.QTimer
    pickups: List[PickupEntry]

    def __init__(self):
        super().__init__()
        self.logger.setLevel(logging.DEBUG)
        self.window = QMainWindow()
        self.setupUi(self.window)
        common_qt_lib.set_default_window_icon(self.window)

        self.pickups = []

        self.collect_location_combo.setVisible(False)
        self.setup_collect_location_combo_button = QtWidgets.QPushButton(self.window)
        self.setup_collect_location_combo_button.setText("Load list of locations")
        self.setup_collect_location_combo_button.clicked.connect(self._setup_locations_combo)
        self.gridLayout.addWidget(self.setup_collect_location_combo_button, 0, 0, 1, 2)

        self.collect_location_button.clicked.connect(self._emit_collection)
        self.collect_location_button.setEnabled(False)
        self.collect_randomly_check.stateChanged.connect(self._on_collect_randomly_toggle)
        self._timer = QtCore.QTimer(self.window)
        self._timer.timeout.connect(self._collect_randomly)
        self._timer.setInterval(10000)

        self._used_version = echoes_dol_versions.ALL_VERSIONS[0]
        self._connector = EchoesRemoteConnector(self._used_version)
        self.game = default_database.game_description_for(RandovaniaGame.METROID_PRIME_ECHOES)

        self._game_memory = bytearray(24 * (2 ** 20))
        self._game_memory_initialized = False

    def _on_collect_randomly_toggle(self, value: int):
        if bool(value):
            self._timer.start()
        else:
            self._timer.stop()

    def _collect_randomly(self):
        row = random.randint(0, self.collect_location_combo.count())
        self._collect_location(self.collect_location_combo.itemData(row))

    async def _ensure_initialized_game_memory(self):
        if self._game_memory_initialized:
            return

        world = self.game.world_list.worlds[0]

        await self.perform_memory_operations([
            # Build String
            MemoryOperation(self._used_version.build_string_address, write_bytes=self._used_version.build_string),

            # current CWorld
            MemoryOperation(self._used_version.game_state_pointer, offset=4,
                            write_bytes=world.world_asset_id.to_bytes(4, "big")),

            # CPlayer VTable
            MemoryOperation(self._used_version.cstate_manager_global + 0x14fc, offset=0,
                            write_bytes=self._used_version.cplayer_vtable.to_bytes(4, "big")),

            # CPlayerState
            MemoryOperation(self._used_version.cstate_manager_global + 0x150c,
                            write_bytes=0xA00000.to_bytes(4, "big")),
        ])

        self._game_memory_initialized = True

    def _read_memory(self, address: int, count: int):
        address &= ~0x80000000
        return self._game_memory[address:address + count]

    def _read_memory_format(self, format_string: str, address: int):
        return struct.unpack_from(format_string, self._game_memory, address & ~0x80000000)

    def _write_memory(self, address: int, data: bytes):
        address &= ~0x80000000
        self._game_memory[address:address + len(data)] = data
        self.logger.info(f"Wrote {data.hex()} to {hex(address)}")

    async def _update_inventory_label(self):
        inventory = await self._connector.get_inventory(self)

        s = "<br />".join(
            f"{name} x {quantity.amount}/{quantity.capacity}" for name, quantity in inventory.items()
        )
        self.inventory_label.setText(s)

    def show(self):
        self.window.show()

    def _get_magic_address(self):
        multiworld_magic_item = self.game.resource_database.multiworld_magic_item
        return _echoes_powerup_address(multiworld_magic_item.index)

    def _read_magic(self):
        return self._read_memory_format(">II", self._get_magic_address())

    def _write_magic(self, magic_amount, magic_capacity):
        self._write_memory(self._get_magic_address(), struct.pack(">II", magic_amount, magic_capacity))

    def _emit_collection(self):
        self._collect_location(self.collect_location_combo.currentData())

    def _collect_location(self, location: int):
        new_magic_value = location + 1
        magic_amount, magic_capacity = self._read_magic()
        if magic_amount == 0:
            magic_amount += new_magic_value
            magic_capacity += new_magic_value
            self._write_magic(magic_amount, magic_capacity)

    @asyncSlot()
    @handle_network_errors
    async def _setup_locations_combo(self):
        network_client = common_qt_lib.get_network_client()
        game_session = network_client.current_game_session
        user = network_client.current_user

        game = self.game
        index_to_name = {
            node.pickup_index.index: game.world_list.area_name(area)
            for world, area, node in game.world_list.all_worlds_areas_nodes
            if isinstance(node, PickupNode)
        }

        if game_session is None:
            names = index_to_name
        else:
            patcher_data = await network_client.session_admin_player(user.id,
                                                                     SessionAdminUserAction.CREATE_PATCHER_FILE,
                                                                     BaseCosmeticPatches().as_json)
            names = {
                pickup["pickup_index"]: "{}: {}".format(index_to_name[pickup["pickup_index"]],
                                                        pickup["hud_text"][0].split(", ", 1)[0])
                for pickup in patcher_data["pickups"]
            }

        self.collect_location_combo.clear()
        for index, name in sorted(names.items(), key=lambda t: t[1]):
            self.collect_location_combo.addItem(name, index)

        self.collect_location_button.setEnabled(True)
        self.collect_location_combo.setVisible(True)
        self.setup_collect_location_combo_button.deleteLater()

    def clear(self):
        self.messages_list.clear()
        self.pickups.clear()

    def _memory_operation(self, op: MemoryOperation) -> Optional[bytes]:
        op.validate_byte_sizes()

        address = op.address
        if op.offset is not None:
            address = self._read_memory_format(">I", address)[0]
            address += op.offset

        result = None
        if op.read_byte_count is not None:
            result = self._read_memory(address, op.read_byte_count)
        if op.write_bytes is not None:
            self._write_memory(address, op.write_bytes)
        return result

    def _add_power_up(self, registers: Dict[int, int]):
        item_id = registers[4]
        delta = registers[5]
        item = self.game.resource_database.get_item(item_id)
        address = _echoes_powerup_address(item_id)

        amount, capacity = self._read_memory_format(">II", address)
        capacity = max(0, min(capacity + delta, item.max_capacity))
        amount = min(amount, capacity)
        self._write_memory(address, struct.pack(">II", amount, capacity))

    def _incr_pickup(self, registers: Dict[int, int]):
        item_id = registers[4]
        delta = registers[5]
        address = _echoes_powerup_address(item_id)

        amount, capacity = self._read_memory_format(">II", address)
        amount = max(0, min(amount + delta, capacity))
        self._write_memory(address, struct.pack(">I", amount))

    def _decr_pickup(self, registers: Dict[int, int]):
        item_id = registers[4]
        delta = registers[5]
        address = _echoes_powerup_address(item_id)

        amount, capacity = self._read_memory_format(">II", address)
        amount = max(0, min(amount - delta, capacity))
        self._write_memory(address, struct.pack(">I", amount))

    def _display_hud_memo(self, registers):
        string_start = self._used_version.string_display.message_receiver_string_ref
        message_bytes = self._read_memory(string_start, self._used_version.string_display.max_message_size + 2)
        message = message_bytes[:message_bytes.find(b"\x00\x00")].decode("utf-16_be")
        self.messages_list.addItem(message)

    def _handle_remote_execution(self):
        has_message_address = self._used_version.cstate_manager_global + 0x2
        if self._read_memory(has_message_address, 1) == b"\x00":
            return

        self.logger.debug("has code to execute, start parsing ppc")
        registers: Dict[int, int] = {i: 0 for i in range(32)}
        registers[3] = self._used_version.sda2_base

        functions = {
            self._used_version.powerup_functions.add_power_up: self._add_power_up,
            self._used_version.powerup_functions.incr_pickup: self._incr_pickup,
            self._used_version.powerup_functions.decr_pickup: self._decr_pickup,
            self._used_version.string_display.wstring_constructor: lambda reg: None,
            self._used_version.string_display.display_hud_memo: self._display_hud_memo,
        }

        patch_address, _ = all_prime_dol_patches.create_remote_execution_body(self._used_version.string_display, [])
        current_address = patch_address
        while current_address - patch_address < 0x2000:
            instruction = self._read_memory(current_address, 4)
            inst_value = int.from_bytes(instruction, "big")
            header = inst_value >> 26

            def get_signed_16():
                return struct.unpack("h", struct.pack("H", inst_value & 65535))[0]

            if header == 19:
                # subset of branch without address. Consider these blr
                end_address = current_address
                self.logger.debug("blr")
                break

            if header == 14:
                # addi
                output_reg = (inst_value >> 21) & 31
                input_reg = (inst_value >> 16) & 31
                value = get_signed_16()
                if input_reg == 0:
                    self.logger.debug(f"addi r{output_reg}, {value}")
                    registers[output_reg] = value
                else:
                    self.logger.debug(f"addi r{output_reg}, r{input_reg}, {value}")
                    registers[output_reg] = registers[input_reg] + value

            elif header == 15:
                # addis
                output_reg = (inst_value >> 21) & 31
                input_reg = (inst_value >> 16) & 31
                value = inst_value & 65535
                if input_reg == 0:
                    self.logger.debug(f"lis r{output_reg}, {value}")
                    registers[output_reg] = value << 16
                else:
                    self.logger.debug(f"addis r{output_reg}, r{input_reg}, {value}")
                    registers[output_reg] = (registers[input_reg] + value) << 16

            elif header == 18:
                # b and bl (relative branch)
                jump_offset = (inst_value >> 2) & ((2 << 24) - 1)
                link = bool(inst_value & 1)
                address = current_address + (jump_offset * 4)

                if address in functions:
                    function_name = functions[address].__name__
                    functions[address](registers)
                else:
                    function_name = "unknown"

                self.logger.debug(f"b{'l' if link else ''} 0x{address:0x} [{function_name}]")

            elif header == 24:
                # ori
                output_reg = (inst_value >> 21) & 31
                input_reg = (inst_value >> 16) & 31
                value = inst_value & 65535

                self.logger.debug(f"ori r{output_reg}, r{input_reg}, {value}")
                registers[output_reg] = registers[input_reg] | value

            elif header == 36 or header == 38:
                # stw (36) and stb (38)
                input_reg = (inst_value >> 21) & 31
                output_reg = (inst_value >> 16) & 31
                value = get_signed_16()

                if header == 36:
                    inst = 'stw'
                    size = 4
                else:
                    inst = 'stb'
                    size = 1

                self.logger.debug(f"{inst} r{input_reg}, 0x{value:0x} (r{output_reg})")
                # ignore r1, as it's writing to stack
                if output_reg != 1:
                    self._write_memory(registers[output_reg] + value,
                                       registers[input_reg].to_bytes(size, "big"))

            elif header == 32:
                # lwz
                output_reg = (inst_value >> 21) & 31
                input_reg = (inst_value >> 16) & 31
                value = get_signed_16()

                self.logger.debug(f"lwz r{input_reg}, 0x{value:0x} (r{output_reg})")
                ea = value
                if input_reg != 0:
                    ea += registers[input_reg]

                registers[output_reg] = int.from_bytes(self._read_memory(ea, 4), "big")
            else:
                self.logger.debug(f"unsupported instruction: {instruction.hex()}; {header}")

            current_address += 4

        self._write_memory(has_message_address, b"\x00")

    # MemoryOperationExecutor

    @property
    def lock_identifier(self) -> Optional[str]:
        return None

    @property
    def backend_choice(self):
        return DebugGameBackendChoice()

    async def connect(self) -> bool:
        await self._ensure_initialized_game_memory()
        self._handle_remote_execution()
        await self._update_inventory_label()
        self._connected = True
        return True

    async def disconnect(self):
        self._connected = False

    def is_connected(self) -> bool:
        return self._connected

    async def perform_memory_operations(self, ops: List[MemoryOperation]) -> Dict[MemoryOperation, bytes]:
        result = {}
        for op in ops:
            op_result = self._memory_operation(op)
            if op_result is not None:
                result[op] = op_result
        return result



import dataclasses
import uuid

import pytest

from randovania.games.game import RandovaniaGame
from randovania.gui.preset_settings.logic_settings_window import LogicSettingsWindow
from randovania.interface_common.preset_editor import PresetEditor


@pytest.mark.parametrize("game", [RandovaniaGame.METROID_PRIME, RandovaniaGame.METROID_PRIME_ECHOES, RandovaniaGame.METROID_PRIME_CORRUPTION])
def test_on_preset_changed(skip_qtbot, preset_manager, game):
    # Setup
    base = preset_manager.default_preset_for_game(game).get_preset()
    preset = dataclasses.replace(base,
                                 uuid=uuid.UUID('b41fde84-1f57-4b79-8cd6-3e5a78077fa6'),
                                 base_preset_uuid=base.uuid)
    editor = PresetEditor(preset)
    window = LogicSettingsWindow(None, editor)

    # Run
    window.on_preset_changed(editor.create_custom_preset_with())



import dataclasses
from enum import Enum
from typing import Dict, List

from randovania.bitpacking.bitpacking import BitPackEnum, BitPackDataclass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.bitpacking.type_enforcement import DataclassPostInitTypeCheck
from randovania.game_description import default_database
from randovania.game_description.world.area import Area
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.games.game import RandovaniaGame
from randovania.layout.lib import location_list


class TeleporterShuffleMode(BitPackEnum, Enum):
    VANILLA = "vanilla"
    TWO_WAY_RANDOMIZED = "randomized"
    TWO_WAY_UNCHECKED = "two-way-unchecked"
    ONE_WAY_ELEVATOR = "one-way-elevator"
    ONE_WAY_ELEVATOR_REPLACEMENT = "one-way-elevator-replacement"
    ONE_WAY_ANYTHING = "one-way-anything"

    @property
    def long_name(self) -> str:
        if self == TeleporterShuffleMode.VANILLA:
            return "Original connections"
        elif self == TeleporterShuffleMode.TWO_WAY_RANDOMIZED:
            return "Two-way, between areas"
        elif self == TeleporterShuffleMode.TWO_WAY_UNCHECKED:
            return "Two-way, unchecked"
        elif self == TeleporterShuffleMode.ONE_WAY_ELEVATOR:
            return "One-way, elevator room with cycles"
        elif self == TeleporterShuffleMode.ONE_WAY_ELEVATOR_REPLACEMENT:
            return "One-way, elevator room with replacement"
        elif self == TeleporterShuffleMode.ONE_WAY_ANYTHING:
            return "One-way, anywhere"
        else:
            raise ValueError(f"Unknown value: {self}")


def _has_editable_teleporter(area: Area) -> bool:
    return any(
        node.editable
        for node in area.nodes
        if isinstance(node, TeleporterNode)
    )


class TeleporterList(location_list.LocationList):
    @classmethod
    def areas_list(cls, game: RandovaniaGame) -> List[Teleporter]:
        world_list = default_database.game_description_for(game).world_list
        areas = [
            node.teleporter
            for world in world_list.worlds
            for area in world.areas
            for node in area.nodes
            if isinstance(node, TeleporterNode) and node.editable
        ]
        areas.sort()
        return areas

    @classmethod
    def element_type(cls):
        return Teleporter

    def ensure_has_location(self, area_location: Teleporter, enabled: bool) -> "TeleporterList":
        return super().ensure_has_location(area_location, enabled)

    def ensure_has_locations(self, area_locations: List[AreaLocation], enabled: bool) -> "TeleporterList":
        return super().ensure_has_locations(area_locations, enabled)


def _valid_teleporter_target(area: Area, game: RandovaniaGame):
    if game == RandovaniaGame.METROID_PRIME and area.area_asset_id == 3031702600:
        return True

    if game == RandovaniaGame.METROID_PRIME_ECHOES and area.area_asset_id == 1393588666:
        return True

    has_save_station = any(node.name == "Save Station" for node in area.nodes)
    return area.valid_starting_location and area.default_node_index is not None and not has_save_station


class TeleporterTargetList(location_list.LocationList):
    @classmethod
    def areas_list(cls, game: RandovaniaGame):
        return location_list.area_locations_with_filter(game, lambda area: _valid_teleporter_target(area, game))


@dataclasses.dataclass(frozen=True)
class TeleporterConfiguration(BitPackDataclass, JsonDataclass, DataclassPostInitTypeCheck):
    mode: TeleporterShuffleMode
    excluded_teleporters: TeleporterList
    excluded_targets: TeleporterTargetList
    skip_final_bosses: bool
    allow_unvisited_room_names: bool

    @property
    def game(self) -> RandovaniaGame:
        return self.excluded_teleporters.game

    @property
    def is_vanilla(self):
        return self.mode == TeleporterShuffleMode.VANILLA

    @property
    def can_use_unvisited_room_names(self) -> bool:
        return self.is_vanilla or self.allow_unvisited_room_names

    @property
    def has_shuffled_target(self):
        return self.mode == TeleporterShuffleMode.ONE_WAY_ANYTHING

    @property
    def editable_teleporters(self) -> List[Teleporter]:
        return [teleporter for teleporter in self.excluded_teleporters.areas_list(self.game)
                if teleporter not in self.excluded_teleporters.locations]

    @property
    def static_teleporters(self) -> Dict[Teleporter, AreaLocation]:
        static = {}
        if self.skip_final_bosses:
            if self.game == RandovaniaGame.METROID_PRIME:
                static[Teleporter(972217896, 597223686, 1049306)] = AreaLocation(0x13d79165, 0xb4b41c48)
            elif self.game == RandovaniaGame.METROID_PRIME_ECHOES:
                static[Teleporter(1006255871, 2278776548, 136970379)] = AreaLocation(1006255871, 1393588666)
            else:
                raise ValueError(f"Unsupported skip_final_bosses and {self.game}")

        return static

    @property
    def valid_targets(self) -> List[AreaLocation]:
        if self.mode == TeleporterShuffleMode.ONE_WAY_ANYTHING:
            return [location for location in self.excluded_targets.areas_list(self.game)
                    if location not in self.excluded_targets.locations]

        elif self.mode in {TeleporterShuffleMode.ONE_WAY_ELEVATOR, TeleporterShuffleMode.ONE_WAY_ELEVATOR_REPLACEMENT}:
            world_list = default_database.game_description_for(self.game).world_list
            return [
                world_list.teleporter_to_node(teleporter).default_connection
                for teleporter in self.editable_teleporters
            ]
        else:
            return []

    def description(self):
        result = []
        if not self.is_vanilla and self.excluded_teleporters.locations:
            result.append(f"{len(self.excluded_teleporters.locations)} teleporters")

        if self.has_shuffled_target and self.excluded_targets.locations:
            result.append(f"{len(self.excluded_targets.locations)} targets")

        if result:
            return f"{self.mode.long_name}; excluded {', '.join(result)}"
        else:
            return self.mode.long_name

    def dangerous_settings(self):
        if self.mode == TeleporterShuffleMode.ONE_WAY_ANYTHING:
            return ["One-way anywhere elevators"]
        return []



import random
from random import Random
from typing import List
from unittest.mock import patch, MagicMock

import pytest

from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.generator import elevator_distributor
from randovania.generator.elevator_distributor import ElevatorHelper


@pytest.mark.parametrize(["seed_number", "expected_ids"], [
    (5000, [122, 129, 1638535, 393260, 4522032, 204865660, 4260106, 2097251, 38, 589851,
            2162826, 1245332, 1572998, 1245307, 3342446, 524321, 2949235, 1966093, 3538975, 152]),
    (9000, [2949235, 129, 152, 4522032, 4260106, 1245332, 1966093, 122, 1638535, 393260,
            204865660, 589851, 1572998, 38, 2097251, 2162826, 3538975, 524321, 1245307, 3342446]),
    (1157772449, [393260, 204865660, 4522032, 122, 129, 2949235, 1638535, 589851, 38, 2097251,
                  4260106, 3538975, 3342446, 1245332, 1966093, 524321, 2162826, 152, 1572998, 1245307])
])
def test_try_randomize_elevators(seed_number: int,
                                 expected_ids: List[int],
                                 echoes_game_description):
    # Setup
    rng = Random(seed_number)
    teleporters = [
        node.teleporter
        for world in echoes_game_description.world_list.worlds
        for area in world.areas
        for node in area.nodes
        if isinstance(node, TeleporterNode) and node.editable and node.teleporter_instance_id in expected_ids
    ]
    teleporters.sort()

    # Run
    result = elevator_distributor.try_randomize_elevators(
        rng,
        elevator_distributor.create_elevator_database(echoes_game_description.world_list, teleporters))
    connected_ids = [elevator.connected_elevator.instance_id for elevator in result]

    # Assert
    assert connected_ids == expected_ids


@patch("randovania.generator.elevator_distributor.try_randomize_elevators", autospec=True)
def test_two_way_elevator_connections_between_areas(mock_try_randomize_elevators: MagicMock,
                                                    ):
    # Setup
    rng = MagicMock()
    elevator_a = MagicMock()
    elevator_b = MagicMock()
    mock_try_randomize_elevators.return_value = [
        elevator_a, elevator_b
    ]

    # Run
    result = elevator_distributor.two_way_elevator_connections(rng, (elevator_a, elevator_b), True)

    # Assert
    mock_try_randomize_elevators.assert_called_once_with(rng, (elevator_a, elevator_b))
    assert result == {
        elevator_a.teleporter: elevator_a.connected_elevator.area_location,
        elevator_b.teleporter: elevator_b.connected_elevator.area_location,
    }


def test_two_way_elevator_connections_unchecked():
    # Setup
    rng = random.Random(5000)
    elevators = [
        ElevatorHelper(Teleporter(i, i, i), AreaLocation(i, i))
        for i in range(6)
    ]
    database = tuple(elevators)

    # Run
    result = elevator_distributor.two_way_elevator_connections(rng, database, False)

    # Assert
    assert result == {
        Teleporter(0, 0, 0): AreaLocation(4, 4),
        Teleporter(1, 1, 1): AreaLocation(2, 2),
        Teleporter(2, 2, 2): AreaLocation(1, 1),
        Teleporter(3, 3, 3): AreaLocation(5, 5),
        Teleporter(4, 4, 4): AreaLocation(0, 0),
        Teleporter(5, 5, 5): AreaLocation(3, 3),
    }


@pytest.mark.parametrize(["replacement", "expected"], [
    (False, {
        Teleporter(0, 0, 0): AreaLocation(1, 1),
        Teleporter(1, 1, 1): AreaLocation(2, 2),
        Teleporter(2, 2, 2): AreaLocation(3, 3),
        Teleporter(3, 3, 3): AreaLocation(5, 5),
        Teleporter(4, 4, 4): AreaLocation(0, 0),
        Teleporter(5, 5, 5): AreaLocation(4, 4),
    }),
    (True, {
        Teleporter(0, 0, 0): AreaLocation(2, 2),
        Teleporter(1, 1, 1): AreaLocation(3, 3),
        Teleporter(2, 2, 2): AreaLocation(4, 4),
        Teleporter(3, 3, 3): AreaLocation(2, 2),
        Teleporter(4, 4, 4): AreaLocation(5, 5),
        Teleporter(5, 5, 5): AreaLocation(3, 3),
    }),
])
def test_one_way_elevator_connections(echoes_game_description, replacement, expected):
    # Setup
    rng = random.Random(5000)
    target_locations = [
        AreaLocation(i, i)
        for i in range(6)
    ]
    elevators = [
        ElevatorHelper(Teleporter(i, i, i), AreaLocation(i, i))
        for i in range(6)
    ]
    database = tuple(elevators)

    # Run
    result = elevator_distributor.one_way_elevator_connections(rng, database, target_locations, replacement)

    # Assert
    assert result == expected



from typing import Callable, Union

import pytest
from mock import MagicMock, patch, call, AsyncMock

import randovania
from randovania.generator import generator
from randovania.generator.filler.runner import FillerPlayerResult
from randovania.layout.layout_description import LayoutDescription
from randovania.resolver.exceptions import InvalidConfiguration


@patch("randovania.generator.generator._validate_item_pool_size", autospec=True)
@patch("randovania.generator.generator.create_player_pool", autospec=True)
@patch("randovania.generator.generator._distribute_remaining_items", autospec=True)
@patch("randovania.generator.generator.Random", autospec=False)
@pytest.mark.asyncio
async def test_create_patches(mock_random: MagicMock,
                              mock_distribute_remaining_items: MagicMock,
                              mock_create_player_pool: MagicMock,
                              mock_validate_item_pool_size: MagicMock,
                              mocker,
                              ):
    # Setup
    filler_result = MagicMock()
    mock_run_filler: AsyncMock = mocker.patch("randovania.generator.generator.run_filler", new_callable=AsyncMock,
                                              return_value=filler_result)

    num_players = 1
    rng = mock_random.return_value
    status_update: Union[MagicMock, Callable[[str], None]] = MagicMock()
    player_pools = [MagicMock() for _ in range(num_players)]
    presets = [MagicMock() for _ in range(num_players)]

    mock_create_player_pool.side_effect = player_pools

    permalink = MagicMock()
    permalink.get_preset.side_effect = lambda i: presets[i]

    # Run
    result = await generator._create_description(permalink, status_update, 0)

    # Assert
    mock_random.assert_called_once_with(permalink.as_bytes)
    mock_create_player_pool.assert_has_calls([
        call(rng, presets[i].configuration, i, num_players)
        for i in range(num_players)
    ])
    mock_validate_item_pool_size.assert_has_calls([
        call(player_pools[i].pickups, player_pools[i].game, player_pools[i].configuration)
        for i in range(num_players)
    ])
    mock_run_filler.assert_awaited_once_with(rng, {i: player_pools[i] for i in range(num_players)}, status_update)
    mock_distribute_remaining_items.assert_called_once_with(rng, filler_result.player_results)

    assert result == LayoutDescription(
        permalink=permalink,
        version=randovania.VERSION,
        all_patches=mock_distribute_remaining_items.return_value,
        item_order=filler_result.action_log,
    )


def test_distribute_remaining_items_no_locations_left(echoes_game_description):
    # Setup
    rng = MagicMock()
    player_result = FillerPlayerResult(
        game=echoes_game_description,
        patches=echoes_game_description.create_game_patches(),
        unassigned_pickups=[MagicMock()] * 1000,
    )
    filler_results = {0: player_result}

    # Run
    with pytest.raises(InvalidConfiguration,
                       match=r"Received 1000 remaining pickups, but there's only \d+ unassigned locations."):
        generator._distribute_remaining_items(rng, filler_results)



from randovania.games.game import RandovaniaGame
from randovania.games.patcher import Patcher


class PatcherProvider:
    def __init__(self):
        from randovania.games.patchers import claris_patcher
        from randovania.games.patchers import randomprime_patcher
        from randovania.games.patchers import super_duper_metroid_patcher
        self._patchers = {
            RandovaniaGame.METROID_PRIME: randomprime_patcher.RandomprimePatcher(),
            RandovaniaGame.METROID_PRIME_ECHOES: claris_patcher.ClarisPatcher(),
            RandovaniaGame.SUPER_METROID: super_duper_metroid_patcher.SuperDuperMetroidPatcher(),
        }

    def patcher_for_game(self, game: RandovaniaGame) -> Patcher:
        return self._patchers[game]



from enum import Enum
from typing import List

from randovania.bitpacking.bitpacking import BitPackEnum


class LayoutTrickLevel(BitPackEnum, Enum):
    DISABLED = "disabled"
    BEGINNER = "beginner"
    INTERMEDIATE = "intermediate"
    ADVANCED = "advanced"
    EXPERT = "expert"
    HYPERMODE = "hypermode"

    @classmethod
    def default(cls) -> "LayoutTrickLevel":
        return cls.DISABLED

    @classmethod
    def maximum(cls) -> "LayoutTrickLevel":
        return _TRICK_LEVEL_ORDER[-1]

    @classmethod
    def from_number(cls, number: int) -> "LayoutTrickLevel":
        return _TRICK_LEVEL_ORDER[number]

    @property
    def as_number(self) -> int:
        return _TRICK_LEVEL_ORDER.index(self)

    @property
    def long_name(self) -> str:
        return _PRETTY_TRICK_LEVEL_NAME[self]


_TRICK_LEVEL_ORDER: List[LayoutTrickLevel] = list(LayoutTrickLevel)
_PRETTY_TRICK_LEVEL_NAME = {
    LayoutTrickLevel.DISABLED: "Disabled",
    LayoutTrickLevel.BEGINNER: "Beginner",
    LayoutTrickLevel.INTERMEDIATE: "Intermediate",
    LayoutTrickLevel.ADVANCED: "Advanced",
    LayoutTrickLevel.EXPERT: "Expert",
    LayoutTrickLevel.HYPERMODE: "Hypermode",
}



import pytest
from randovania.gui.lib.foldable import Foldable


def test_foldable_initial_state(skip_qtbot):
    # Setup & Run
    foldable = Foldable("My foldable title", False)
    foldable_2 = Foldable("")
    
    skip_qtbot.addWidget(foldable)
    skip_qtbot.addWidget(foldable_2)

    # Assert
    assert foldable._toggleButton.text() == "My foldable title"
    assert not foldable._folded
    assert foldable_2._folded

def test_foldable_actions(skip_qtbot):
    # Setup
    foldable = Foldable("My foldable title")

    skip_qtbot.addWidget(foldable)

    # Run & Assert
    foldable._unfold()
    assert not foldable._folded
    assert not foldable._contentArea.isHidden()
    
    foldable._fold()
    assert foldable._folded
    assert foldable._contentArea.isHidden()


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_patches.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetEchoesPatches(object):
    def setupUi(self, PresetEchoesPatches):
        if not PresetEchoesPatches.objectName():
            PresetEchoesPatches.setObjectName(u"PresetEchoesPatches")
        PresetEchoesPatches.resize(466, 454)
        self.centralWidget = QWidget(PresetEchoesPatches)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.include_menu_mod_check = QCheckBox(self.centralWidget)
        self.include_menu_mod_check.setObjectName(u"include_menu_mod_check")
        self.include_menu_mod_check.setEnabled(True)

        self.verticalLayout.addWidget(self.include_menu_mod_check)

        self.include_menu_mod_label = QLabel(self.centralWidget)
        self.include_menu_mod_label.setObjectName(u"include_menu_mod_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.include_menu_mod_label.sizePolicy().hasHeightForWidth())
        self.include_menu_mod_label.setSizePolicy(sizePolicy)
        self.include_menu_mod_label.setMaximumSize(QSize(16777215, 60))
        self.include_menu_mod_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.include_menu_mod_label.setWordWrap(True)
        self.include_menu_mod_label.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.include_menu_mod_label)

        self.line_1 = QFrame(self.centralWidget)
        self.line_1.setObjectName(u"line_1")
        self.line_1.setFrameShape(QFrame.HLine)
        self.line_1.setFrameShadow(QFrame.Sunken)

        self.verticalLayout.addWidget(self.line_1)

        self.warp_to_start_check = QCheckBox(self.centralWidget)
        self.warp_to_start_check.setObjectName(u"warp_to_start_check")

        self.verticalLayout.addWidget(self.warp_to_start_check)

        self.warp_to_start_label = QLabel(self.centralWidget)
        self.warp_to_start_label.setObjectName(u"warp_to_start_label")
        sizePolicy.setHeightForWidth(self.warp_to_start_label.sizePolicy().hasHeightForWidth())
        self.warp_to_start_label.setSizePolicy(sizePolicy)
        self.warp_to_start_label.setMaximumSize(QSize(16777215, 70))
        self.warp_to_start_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.warp_to_start_label.setWordWrap(True)

        self.verticalLayout.addWidget(self.warp_to_start_label)

        PresetEchoesPatches.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesPatches)

        QMetaObject.connectSlotsByName(PresetEchoesPatches)
    # setupUi

    def retranslateUi(self, PresetEchoesPatches):
        PresetEchoesPatches.setWindowTitle(QCoreApplication.translate("PresetEchoesPatches", u"Other", None))
        self.include_menu_mod_check.setText(QCoreApplication.translate("PresetEchoesPatches", u"Include Menu Mod", None))
        self.include_menu_mod_label.setText(QCoreApplication.translate("PresetEchoesPatches", u"<html><head/><body><p>Menu Mod is a practice tool for Echoes, allowing in-game changes to which items you have and warping to all rooms.</p><p>See the <a href=\"https://www.dropbox.com/s/yhqqafaxfo3l4vn/Echoes%20Menu.7z?file_subpath=%2FEchoes+Menu%2Freadme.txt\"><span style=\" text-decoration: underline; color:#0000ff;\">Menu Mod README</span></a> for more details.</p></body></html>", None))
        self.warp_to_start_check.setText(QCoreApplication.translate("PresetEchoesPatches", u"Add warping to starting location from save stations", None))
        self.warp_to_start_label.setText(QCoreApplication.translate("PresetEchoesPatches", u"<html><head/><body><p>Refusing to save at any Save Station will prompt if you want to warp to the starting location (by default, Samus' ship in Landing Site).</p><p><span style=\" color:#005500;\">Usage of the this option is encouraged for all, as it prevents many softlocks that occurs normally in Echoes.</span></p></body></html>", None))
    # retranslateUi




import dataclasses
import json
import logging
from enum import Enum
from typing import Dict, List, Union, Optional

import PySide2
from PySide2 import QtWidgets
from PySide2.QtCore import QTimer, Signal, Qt
from PySide2.QtGui import QPixmap
from PySide2.QtWidgets import QMainWindow, QLabel, QSpacerItem, QSizePolicy, QActionGroup
from qasync import asyncSlot

from randovania import get_data_path
from randovania.game_connection.connection_base import GameConnectionStatus, InventoryItem
from randovania.game_connection.game_connection import GameConnection
from randovania.game_description import default_database
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.search import find_resource_info_with_long_name
from randovania.games.game import RandovaniaGame
from randovania.gui.generated.auto_tracker_window_ui import Ui_AutoTrackerWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.clickable_label import ClickableLabel
from randovania.gui.lib.game_connection_setup import GameConnectionSetup
from randovania.gui.lib.pixmap_lib import paint_with_opacity
from randovania.interface_common.options import Options


def path_for(name: str):
    return get_data_path().joinpath(f"gui_assets/tracker", name)


class FieldToCheck(Enum):
    AMOUNT = "amount"
    CAPACITY = "capacity"
    MAX_CAPACITY = "max_capacity"


@dataclasses.dataclass(frozen=True)
class Element:
    labels: List[Union[QLabel, ClickableLabel, ClickableLabel]]
    resources: List[ItemResourceInfo]
    text_template: str
    minimum_to_check: int
    field_to_check: FieldToCheck

    def __post_init__(self):
        if len(self.labels) > 1 and len(self.labels) != len(self.resources):
            raise ValueError("Label has {} progressive icons, but has {} resources ({}).".format(
                len(self.labels),
                len(self.resources),
                str([r.long_name for r in self.resources])
            ))


class AutoTrackerWindow(QMainWindow, Ui_AutoTrackerWindow):
    _hooked = False
    _base_address: int
    trackers: Dict[str, str]
    _current_tracker_game: RandovaniaGame = None
    _current_tracker_name: str = None
    give_item_signal = Signal(PickupEntry)

    def __init__(self, game_connection: GameConnection, options: Options):
        super().__init__()
        self.setupUi(self)
        self.game_connection = game_connection
        self.options = options
        common_qt_lib.set_default_window_icon(self)

        with get_data_path().joinpath(f"gui_assets/tracker/trackers.json").open("r") as trackers_file:
            self.trackers = json.load(trackers_file)["trackers"]

        self._action_to_name = {}
        theme_group = QActionGroup(self)
        for name in self.trackers.keys():
            action = QtWidgets.QAction(self.menu_tracker)
            action.setText(name)
            action.setCheckable(True)
            action.setChecked(name == options.selected_tracker)
            action.triggered.connect(self._on_action_select_tracker)
            self.menu_tracker.addAction(action)
            self._action_to_name[action] = name
            theme_group.addAction(action)

        self._tracker_elements: List[Element] = []
        self.create_tracker()

        self.game_connection_setup = GameConnectionSetup(self, self.connection_status_label,
                                                         self.game_connection, options)
        self.game_connection_setup.create_backend_entries(self.menu_backend)
        self.game_connection_setup.create_upload_nintendont_action(self.menu_options)
        self.game_connection_setup.refresh_backend()

        self.action_force_update.triggered.connect(self.on_force_update_button)

        self._update_timer = QTimer(self)
        self._update_timer.setInterval(100)
        self._update_timer.timeout.connect(self._on_timer_update)
        self._update_timer.setSingleShot(True)

    def showEvent(self, event: PySide2.QtGui.QShowEvent):
        self._update_timer.start()
        super().showEvent(event)

    def hideEvent(self, event: PySide2.QtGui.QHideEvent):
        self._update_timer.stop()
        super().hideEvent(event)

    @property
    def selected_tracker(self) -> Optional[str]:
        for action, name in self._action_to_name.items():
            if action.isChecked():
                return name

    def _update_tracker_from_hook(self, inventory: Dict[ItemResourceInfo, InventoryItem]):
        for element in self._tracker_elements:
            if len(element.labels) > 1:
                satisfied = False
                for i, resource in reversed(list(enumerate(element.resources))):
                    current = inventory.get(resource, InventoryItem(0, 0))
                    fields = {"amount": current.amount, "capacity": current.capacity,
                              "max_capacity": resource.max_capacity}

                    if satisfied:
                        element.labels[i].setVisible(False)

                    elif fields[element.field_to_check.value] >= element.minimum_to_check:
                        # This tier is satisfied
                        satisfied = True
                        element.labels[i].setVisible(True)
                        element.labels[i].set_checked(True)
                    else:
                        element.labels[i].setVisible(False)

                if not satisfied:
                    element.labels[0].setVisible(True)
                    element.labels[0].set_checked(False)

            else:
                label = element.labels[0]

                amount = 0
                capacity = 0
                max_capacity = 0
                for resource in element.resources:
                    current = inventory.get(resource, InventoryItem(0, 0))
                    amount += current.amount
                    capacity += current.capacity
                    max_capacity += resource.max_capacity

                if isinstance(label, ClickableLabel):
                    fields = {"amount": amount, "capacity": capacity, "max_capacity": max_capacity}
                    value_target = element.minimum_to_check
                    value = fields[element.field_to_check.value]
                    label.set_checked(max_capacity == 0 or value >= value_target)
                else:
                    label.setText(element.text_template.format(
                        amount=amount,
                        capacity=capacity,
                        max_capacity=max_capacity,
                    ))

    def _on_action_select_tracker(self):
        with self.options as options:
            options.selected_tracker = self.selected_tracker
        self.create_tracker()

    @asyncSlot()
    async def _on_timer_update(self):
        try:
            current_status = self.game_connection.current_status
            if current_status not in {GameConnectionStatus.Disconnected, GameConnectionStatus.UnknownGame,
                                      GameConnectionStatus.WrongGame}:
                self.action_force_update.setEnabled(True)
            else:
                self.action_force_update.setEnabled(False)

            if current_status == GameConnectionStatus.InGame or current_status == GameConnectionStatus.TrackerOnly:
                if self.game_connection.connector.game_enum == self._current_tracker_game:
                    inventory = self.game_connection.get_current_inventory()
                    self._update_tracker_from_hook(inventory)
                    self.game_connection_setup.on_game_connection_updated()
                else:
                    self.connection_status_label.setText("{}: Wrong Game ({})".format(
                        self.game_connection.backend_choice.pretty_text,
                        self.game_connection.current_game_name,
                    ))
        finally:
            self._update_timer.start()

    def delete_tracker(self):
        for element in self._tracker_elements:
            for l in element.labels:
                l.deleteLater()

        self._tracker_elements.clear()

    def create_tracker(self):
        tracker_name = self.selected_tracker
        if tracker_name == self._current_tracker_name or tracker_name is None:
            return

        self.delete_tracker()

        with path_for(self.trackers[tracker_name]).open("r") as tracker_details_file:
            tracker_details = json.load(tracker_details_file)

        game_enum = RandovaniaGame(tracker_details["game"])
        resource_database = default_database.resource_database_for(game_enum)

        for element in tracker_details["elements"]:
            text_template = ""
            minimum_to_check = element.get("minimum_to_check", 1)
            field_to_check = FieldToCheck(element.get("field_to_check", FieldToCheck.CAPACITY.value))

            labels = []
            if "image_path" in element:
                paths = element["image_path"]
                if not isinstance(paths, list):
                    paths = [paths]

                visible = True
                for path in paths:
                    image_path = get_data_path().joinpath(path)
                    if not image_path.exists():
                        logging.error("Tracker asset not found: %s", image_path)
                    pixmap = QPixmap(str(image_path))

                    label = ClickableLabel(self.inventory_group, paint_with_opacity(pixmap, 0.3),
                                           paint_with_opacity(pixmap, 1.0))
                    label.set_checked(False)
                    label.set_ignore_mouse_events(True)
                    label.setVisible(visible)
                    visible = False
                    labels.append(label)

            elif "label" in element:
                label = QLabel(self.inventory_group)
                label.setAlignment(Qt.AlignCenter)
                text_template = element["label"]
                labels.append(label)

            else:
                raise ValueError(f"Invalid element: {element}")

            resources = [
                find_resource_info_with_long_name(resource_database.item, resource_name)
                for resource_name in element["resources"]
            ]
            self._tracker_elements.append(Element(labels, resources, text_template, minimum_to_check, field_to_check))
            for l in labels:
                self.inventory_layout.addWidget(l, element["row"], element["column"])

        self.inventory_spacer = QSpacerItem(5, 5, QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.inventory_layout.addItem(self.inventory_spacer, self.inventory_layout.rowCount(),
                                      self.inventory_layout.columnCount())

        self._current_tracker_game = game_enum
        self._update_tracker_from_hook({})

        self._current_tracker_name = tracker_name

    @asyncSlot()
    async def on_force_update_button(self):
        await self.game_connection.update_current_inventory()
        inventory = self.game_connection.get_current_inventory()
        logging.info("Inventory:" + "\n".join(
            f"{item.long_name}: {inv_item.amount}/{inv_item.capacity}"
            for item, inv_item in sorted(inventory.items(), key=lambda it: it[0].long_name)
        ))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'corruption_cosmetic_patches_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CorruptionCosmeticPatchesDialog(object):
    def setupUi(self, CorruptionCosmeticPatchesDialog):
        if not CorruptionCosmeticPatchesDialog.objectName():
            CorruptionCosmeticPatchesDialog.setObjectName(u"CorruptionCosmeticPatchesDialog")
        CorruptionCosmeticPatchesDialog.resize(396, 246)
        self.gridLayout = QGridLayout(CorruptionCosmeticPatchesDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.reset_button = QPushButton(CorruptionCosmeticPatchesDialog)
        self.reset_button.setObjectName(u"reset_button")

        self.gridLayout.addWidget(self.reset_button, 2, 2, 1, 1)

        self.accept_button = QPushButton(CorruptionCosmeticPatchesDialog)
        self.accept_button.setObjectName(u"accept_button")

        self.gridLayout.addWidget(self.accept_button, 2, 0, 1, 1)

        self.cancel_button = QPushButton(CorruptionCosmeticPatchesDialog)
        self.cancel_button.setObjectName(u"cancel_button")

        self.gridLayout.addWidget(self.cancel_button, 2, 1, 1, 1)

        self.scrollArea = QScrollArea(CorruptionCosmeticPatchesDialog)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 376, 198))
        self.verticalLayout = QVBoxLayout(self.scroll_area_contents)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.game_changes_box = QGroupBox(self.scroll_area_contents)
        self.game_changes_box.setObjectName(u"game_changes_box")
        self.verticalLayout_2 = QVBoxLayout(self.game_changes_box)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.random_door_colors_check = QCheckBox(self.game_changes_box)
        self.random_door_colors_check.setObjectName(u"random_door_colors_check")

        self.verticalLayout_2.addWidget(self.random_door_colors_check)

        self.random_welding_colors_check = QCheckBox(self.game_changes_box)
        self.random_welding_colors_check.setObjectName(u"random_welding_colors_check")

        self.verticalLayout_2.addWidget(self.random_welding_colors_check)

        self.suit_layout = QHBoxLayout()
        self.suit_layout.setSpacing(6)
        self.suit_layout.setObjectName(u"suit_layout")
        self.suit_label = QLabel(self.game_changes_box)
        self.suit_label.setObjectName(u"suit_label")

        self.suit_layout.addWidget(self.suit_label)

        self.suit_combo = QComboBox(self.game_changes_box)
        self.suit_combo.setObjectName(u"suit_combo")

        self.suit_layout.addWidget(self.suit_combo)


        self.verticalLayout_2.addLayout(self.suit_layout)


        self.verticalLayout.addWidget(self.game_changes_box)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.scrollArea.setWidget(self.scroll_area_contents)

        self.gridLayout.addWidget(self.scrollArea, 1, 0, 1, 3)


        self.retranslateUi(CorruptionCosmeticPatchesDialog)

        QMetaObject.connectSlotsByName(CorruptionCosmeticPatchesDialog)
    # setupUi

    def retranslateUi(self, CorruptionCosmeticPatchesDialog):
        CorruptionCosmeticPatchesDialog.setWindowTitle(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Metroid Prime 3 - Cosmetic Options", None))
        self.reset_button.setText(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Reset to Defaults", None))
        self.accept_button.setText(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Accept", None))
        self.cancel_button.setText(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Cancel", None))
        self.game_changes_box.setTitle(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Game Changes", None))
        self.random_door_colors_check.setText(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Random door colors", None))
        self.random_welding_colors_check.setText(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Random welding colors", None))
        self.suit_label.setText(QCoreApplication.translate("CorruptionCosmeticPatchesDialog", u"Initial suit", None))
    # retranslateUi




from dataclasses import dataclass
from typing import Tuple


@dataclass(frozen=True, order=True)
class ItemCategory:
    name: str
    long_name: str
    hint_details: Tuple[str, str]
    is_major: bool
    is_key: bool = False

    @classmethod
    def from_json(cls, name: str, value: dict) -> "ItemCategory":
        return cls(
            name=name,
            long_name=value["long_name"],
            hint_details=tuple(value["hint_details"]),
            is_major=value["is_major"],
            is_key=value["is_key"] if "is_key" in value else False
        )

    @property
    def as_json(self) -> dict:
        result = {
            "long_name": self.long_name,
            "hint_details": list(self.hint_details),
            "is_major": self.is_major,
        }
        if self.is_key:
            result["is_key"] = self.is_key
        return result

    @property
    def general_details(self) -> Tuple[str, str]:
        if self.is_major:
            return "a ", "major upgrade"
        elif self.is_key:
            return "a ", "key"
        else:
            return "an ", "expansion"

    @property
    def is_expansion(self) -> bool:
        return self.name == "expansion"


USELESS_ITEM_CATEGORY = ItemCategory(
    name="useless",
    long_name="",
    hint_details=("an ", "Energy Transfer Module"),
    is_major=False
)

GENERIC_KEY_CATEGORY = ItemCategory(
    name="key",
    long_name="",
    hint_details=("a ", "key"),
    is_major=False
)



from PySide2.QtWidgets import QDialog

from randovania.layout.game_to_class import AnyCosmeticPatches


class BaseCosmeticPatchesDialog(QDialog):
    def connect_signals(self):
        self.accept_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        self.reset_button.clicked.connect(self.reset)

    def reset(self):
        raise NotImplementedError()

    @property
    def cosmetic_patches(self) -> AnyCosmeticPatches:
        raise NotImplementedError()



import copy
import itertools
from heapq import heappush, heappop
from typing import Dict, Iterator, Tuple, Set, Callable

from randovania.game_description.requirements import RequirementSet


class BaseGraph:
    def copy(self):
        raise NotImplementedError()

    def add_node(self, node: int):
        raise NotImplementedError()

    def add_edge(self, previous_node: int, next_node: int, requirement: RequirementSet):
        raise NotImplementedError()

    def remove_edge(self, previous: int, target: int):
        raise NotImplementedError()

    def has_edge(self, previous_node: int, next_node: int) -> bool:
        raise NotImplementedError()

    def __contains__(self, item: int):
        raise NotImplementedError()

    def edges_data(self) -> Iterator[Tuple[int, int, RequirementSet]]:
        raise NotImplementedError()

    def multi_source_dijkstra(self, sources: Set[int], weight: Callable[[int, int, RequirementSet], float]):
        raise NotImplementedError()

    def strongly_connected_components(self) -> Iterator[Set[int]]:
        raise NotImplementedError()


class RandovaniaGraph(BaseGraph):
    edges: Dict[int, Dict[int, RequirementSet]]

    @classmethod
    def new(cls):
        return cls({})

    def __init__(self, edges: Dict[int, Dict[int, RequirementSet]]):
        import networkx
        self.networkx = networkx
        self.edges = edges

    def copy(self):
        return RandovaniaGraph(
            {
                source: copy.copy(data)
                for source, data in self.edges.items()
            },
        )

    def add_node(self, node: int):
        if node not in self.edges:
            self.edges[node] = {}

    def add_edge(self, previous_node: int, next_node: int, requirement: RequirementSet):
        self.edges[previous_node][next_node] = requirement

    def remove_edge(self, previous: int, target: int):
        self.edges[previous].pop(target)

    def has_edge(self, previous_node: int, next_node: int) -> bool:
        return next_node in self.edges.get(previous_node, {})

    def __contains__(self, item: int):
        return item in self.edges

    def edges_data(self):
        for source, data in self.edges.items():
            for target, requirement in data.items():
                yield source, target, requirement

    def multi_source_dijkstra(self, sources: Set[int], weight: Callable[[int, int, RequirementSet], float]):
        paths = {source: [source] for source in sources}  # dictionary of paths
        edges = self.edges

        push = heappush
        pop = heappop

        dist = {}  # dictionary of final distances
        seen = {}
        # fringe is heapq with 3-tuples (distance,c,node)
        # use the count c to avoid comparing nodes (may not be able to)
        c = itertools.count()
        fringe = []
        for source in sources:
            seen[source] = 0
            push(fringe, (0, next(c), source))

        while fringe:
            (d, _, v) = pop(fringe)
            if v in dist:
                continue  # already searched this node.
            dist[v] = d
            for u, e in edges[v].items():
                cost = weight(v, u, e)
                if cost is None:
                    continue
                vu_dist = dist[v] + cost
                if u in dist:
                    u_dist = dist[u]
                    if vu_dist < u_dist:
                        raise ValueError("Contradictory paths found:", "negative weights?")
                elif u not in seen or vu_dist < seen[u]:
                    seen[u] = vu_dist
                    push(fringe, (vu_dist, next(c), u))
                    paths[u] = paths[v] + [u]

        return dist, paths

    def strongly_connected_components(self) -> Iterator[Set[int]]:
        preorder = {}
        lowlink = {}
        scc_found = set()
        scc_queue = []
        i = 0  # Preorder counter
        for source in self.edges.keys():
            if source not in scc_found:
                queue = [source]
                while queue:
                    v = queue[-1]
                    if v not in preorder:
                        i = i + 1
                        preorder[v] = i
                    done = True
                    for w in self.edges[v].keys():
                        if w not in preorder:
                            queue.append(w)
                            done = False
                            break
                    if done:
                        lowlink[v] = preorder[v]
                        for w in self.edges[v].keys():
                            if w not in scc_found:
                                if preorder[w] > preorder[v]:
                                    lowlink[v] = min([lowlink[v], lowlink[w]])
                                else:
                                    lowlink[v] = min([lowlink[v], preorder[w]])
                        queue.pop()
                        if lowlink[v] == preorder[v]:
                            scc = {v}
                            while scc_queue and preorder[scc_queue[-1]] > preorder[v]:
                                k = scc_queue.pop()
                                scc.add(k)
                            scc_found.update(scc)
                            yield scc
                        else:
                            scc_queue.append(v)



import dataclasses
from typing import Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder
from randovania.game_description import default_database
from randovania.game_description.item.ammo import Ammo


@dataclasses.dataclass(frozen=True)
class AmmoState(BitPackValue):
    ammo_count: Tuple[int, ...] = (0,)
    pickup_count: int = 0
    requires_major_item: bool = True

    def check_consistency(self, ammo: Ammo):
        db = default_database.resource_database_for(ammo.game)

        if len(self.ammo_count) != len(ammo.items):
            raise ValueError(f"Ammo state has {len(self.ammo_count)} ammo counts, expected {len(ammo.items)}")

        for count, ammo_index in zip(self.ammo_count, ammo.items):
            ammo_item = db.get_item(ammo_index)
            if not (0 <= count <= ammo_item.max_capacity):
                raise ValueError(f"Ammo count for item {ammo_index} of value {count} is not "
                                 f"in range [0, {ammo_item.max_capacity}].")

        if self.pickup_count < 0:
            raise ValueError(f"Pickup count must be at least 0, got {self.pickup_count}")

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        ammo: Ammo = metadata["ammo"]
        db = default_database.resource_database_for(ammo.game)

        for count, ammo_index in zip(self.ammo_count, ammo.items):
            ammo_item = db.get_item(ammo_index)
            yield from bitpacking.encode_int_with_limits(
                count,
                (ammo_item.max_capacity // 2, ammo_item.max_capacity + 1),
            )

        yield from bitpacking.encode_big_int(self.pickup_count)
        if ammo.unlocked_by is not None:
            yield from bitpacking.encode_bool(self.requires_major_item)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "AmmoState":
        ammo: Ammo = metadata["ammo"]
        db = default_database.resource_database_for(ammo.game)

        # Ammo Count
        ammo_count = []
        for ammo_index in ammo.items:
            ammo_item = db.get_item(ammo_index)
            ammo_count.append(
                bitpacking.decode_int_with_limits(
                    decoder,
                    (ammo_item.max_capacity // 2, ammo_item.max_capacity + 1),
                )
            )

        # Pickup Count
        pickup_count = bitpacking.decode_big_int(decoder)

        # Require Major Item
        requires_major_item = True
        if ammo.unlocked_by is not None:
            requires_major_item = bitpacking.decode_bool(decoder)

        return cls(
            ammo_count=tuple(ammo_count),
            pickup_count=pickup_count,
            requires_major_item=requires_major_item,
        )

    @property
    def as_json(self) -> dict:
        result: dict = {}

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            result[field.name] = value

        result["ammo_count"] = list(result["ammo_count"])

        return result

    @classmethod
    def from_json(cls, value: dict) -> "AmmoState":
        kwargs = {}

        for field in dataclasses.fields(cls):
            if field.name in value:
                kwargs[field.name] = value[field.name]

        if "ammo_count" in kwargs:
            kwargs["ammo_count"] = tuple(kwargs["ammo_count"])

        return cls(**kwargs)



import io
import json
from pathlib import Path

import pytest

from randovania import get_data_path
from randovania.game_description import data_reader
from randovania.games import binary_data


def test_simple_round_trip():
    sample_data = {
        "game": "prime2",
        "resource_database": {
            "items": [],
            "energy_tank_item_index": 0,
            "item_percentage_index": 0,
            "multiworld_magic_item_index": 0,
            "events": [],
            "tricks": [],
            "damage": [],
            "versions": [],
            "misc": [],
            "requirement_template": {},
            "damage_reductions": [],
        },
        "starting_location": {
            "world_asset_id": 1006255871,
            "area_asset_id": 1655756413
        },
        "initial_states": {
            "Default": [
            ]
        },
        "minimal_logic": None,
        "victory_condition": {
            "type": "and",
            "data": {"comment": None, "items": []},
        },
        "dock_weakness_database": {
            "door": [],
            "portal": [],
            "morph_ball": [],
        },
        "worlds": [],
    }

    b = io.BytesIO()
    binary_data.encode(sample_data, b)

    b.seek(0)
    decoded = binary_data.decode(b)

    assert decoded == sample_data


def test_complex_encode(test_files_dir):
    with test_files_dir.joinpath("prime_data_as_json.json").open("r") as data_file:
        data = json.load(data_file)

    b = io.BytesIO()

    # Run
    binary_data.encode(data, b)
    # Whenever the file format changes, we can use the following line to force update our test file
    # test_files_dir.joinpath("prime_data_as_binary.bin").write_bytes(b.getvalue()); assert False

    # Assert
    assert test_files_dir.joinpath("prime_data_as_binary.bin").read_bytes() == b.getvalue()


def test_complex_decode(test_files_dir):
    # Run
    decoded_data = binary_data.decode_file_path(Path(test_files_dir.joinpath("prime_data_as_binary.bin")))

    # Assert
    with test_files_dir.joinpath("prime_data_as_json.json").open("r") as data_file:
        saved_data = json.load(data_file)

    assert decoded_data == saved_data


def _comparable_dict(value):
    if isinstance(value, dict):
        return [
            (key, _comparable_dict(item))
            for key, item in value.items()
        ]

    if isinstance(value, list):
        return [_comparable_dict(item) for item in value]

    return value


def test_full_data_encode_is_equal(game_enum):
    # The json data may be missing if we're running using a Pyinstaller binary
    # Setup
    data_dir = get_data_path().joinpath("json_data", game_enum.value)
    if not data_dir.is_dir() and get_data_path().joinpath("binary_data", f"{game_enum.value}.bin").is_file():
        pytest.skip("Missing json-based data")

    json_database = data_reader.read_split_file(data_dir)

    b = io.BytesIO()
    binary_data.encode(json_database, b)

    b.seek(0)
    decoded_database = binary_data.decode(b)

    # Run
    assert json_database == decoded_database

    comparable_json = _comparable_dict(json_database)
    comparable_binary = _comparable_dict(decoded_database)
    for a, b in zip(comparable_json, comparable_binary):
        assert a == b

    assert comparable_json == comparable_binary


@pytest.mark.parametrize("req", [
    {"type": "or", "data": {"comment": None, "items": []}},
    {"type": "and", "data": {"comment": None, "items": []}},
    {"type": "resource", "data": {"type": 2, "index": 5, "amount": 7, "negate": True}},
    {"type": "template", "data": "Example Template"},
])
def test_encode_requirement_simple(req):
    # Run
    encoded = binary_data.ConstructRequirement.build(req)
    decoded = binary_data.convert_to_raw_python(binary_data.ConstructRequirement.parse(encoded))

    # Assert
    assert req == decoded


def test_encode_requirement_complex():
    # Setup
    req = {
        "type": "and",
        "data": {
            "comment": None,
            "items": [
                {"type": "or", "data": {"comment": None, "items": []}},
                {"type": "and", "data": {"comment": None, "items": []}},
                {"type": "or", "data": {"comment": None, "items": []}},
                {"type": "resource", "data": {"type": 2, "index": 5, "amount": 7, "negate": True}},
                {"type": "or", "data": {"comment": None, "items": []}},
                {"type": "template", "data": "Example Template"},
            ]
        }
    }

    # Run
    encoded = binary_data.ConstructRequirement.build(req)
    decoded = binary_data.convert_to_raw_python(binary_data.ConstructRequirement.parse(encoded))

    # Assert
    assert req == decoded


def test_encode_resource_database():
    # Setup
    resource_database = {
        "items": [],
        "energy_tank_item_index": 0,
        "item_percentage_index": 0,
        "multiworld_magic_item_index": 0,
        "events": [],
        "tricks": [],
        "damage": [],
        "versions": [],
        "misc": [],
        "requirement_template": {
            "Foo": {
                "type": "or",
                "data": {"comment": None, "items": []}
            }
        },
        "damage_reductions": [],
    }
    resource_database["requirement_template"] = list(resource_database["requirement_template"].items())

    # Run
    encoded = binary_data.ConstructResourceDatabase.build(resource_database)

    # Assert
    assert encoded == b'\x00\x00\x00\x00\x00\x00\x01Foo\x00\x02\x00\x00\x00\x00\x01\x00\x01\x00'



import importlib
from pathlib import Path

from PySide2 import QtWidgets, QtGui
from PySide2.QtCore import Qt

import randovania

_current_dark_theme = None


def set_dark_theme(active: bool, compact: bool = False, *, app: QtWidgets.QApplication = None):
    global _current_dark_theme
    if _current_dark_theme == active:
        return

    if app is None:
        app = QtWidgets.QApplication.instance()
    new_palette = QtGui.QPalette(app.palette())

    import qdarktheme
    if active:
        palette = "dark"
    else:
        palette = "light"

    if randovania.is_frozen():
        # HACK: when frozen with pyinstaller, __spec__.origin is None which causes importlib.resources to not find where
        # the stylesheet.qss to be found
        x = Path(__file__).parents[3].joinpath("qdarktheme/dist", palette, "__init__.py")
        importlib.import_module(f"qdarktheme.dist.{palette}").__spec__.origin = x

    style = qdarktheme.load_stylesheet(theme=palette)
    if compact:
        style += """
    QGroupBox {
        padding: 0px;
    }
    QGroupBox::title {
        padding-bottom: 12px;
    }
    QComboBox {
        padding-right: 10px;
    }
    QPushButton {
        min-width: 60px;
    }
    QToolButton {
        border: 1px solid #32414B;
    }
        """
    
    style += "QScrollArea { border: default; }"

    if active:
        new_palette.setColor(QtGui.QPalette.Link, Qt.cyan)
        new_palette.setColor(QtGui.QPalette.LinkVisited, Qt.cyan)
    else:
        new_palette.setColor(QtGui.QPalette.Link, Qt.blue)

    _current_dark_theme = active
    app.setStyleSheet(style)
    app.setPalette(new_palette)



import copy
import json
import logging
import os
import typing
from pathlib import Path
from random import Random
from typing import Optional, List, Union

import py_randomprime

import randovania
from randovania.dol_patching import assembler
from randovania.game_description import default_database
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.node import PickupNode, TeleporterNode
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.games.patcher import Patcher
from randovania.games.prime import prime1_elevators, all_prime_dol_patches, prime_items
from randovania.games.prime.patcher_file_lib import pickup_exporter, item_names, guaranteed_item_hint, hint_lib, \
    credits_spoiler
from randovania.generator.item_pool import pickup_creator
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.prime1.prime_configuration import PrimeConfiguration
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches
from randovania.lib.status_update_lib import ProgressUpdateCallable

_EASTER_EGG_SHINY_MISSILE = 1024

_STARTING_ITEM_NAME_TO_INDEX = {
    "powerBeam": 0,
    "ice": 1,
    "wave": 2,
    "plasma": 3,
    "missiles": 4,
    "scanVisor": 5,
    "bombs": 6,
    "powerBombs": 7,
    "flamethrower": 8,
    "thermalVisor": 9,
    "charge": 10,
    "superMissile": 11,
    "grapple": 12,
    "xray": 13,
    "iceSpreader": 14,
    "spaceJump": 15,
    "morphBall": 16,
    "combatVisor": 17,
    "boostBall": 18,
    "spiderBall": 19,
    "gravitySuit": 21,
    "variaSuit": 22,
    "phazonSuit": 23,
    "energyTanks": 24,
    "wavebuster": 28
}

# "Power Suit": 20,
# "Combat Visor": 17,
# "Unknown Item 1": 25,
# "Health Refill": 26,
# "Unknown Item 2": 27,

_MODEL_MAPPING = {
    (RandovaniaGame.METROID_PRIME_ECHOES, "ChargeBeam INCOMPLETE"): "Charge Beam",
    (RandovaniaGame.METROID_PRIME_ECHOES, "SuperMissile"): "Super Missile",
    (RandovaniaGame.METROID_PRIME_ECHOES, "ScanVisor INCOMPLETE"): "Scan Visor",
    (RandovaniaGame.METROID_PRIME_ECHOES, "VariaSuit INCOMPLETE"): "Varia Suit",
    (RandovaniaGame.METROID_PRIME_ECHOES, "DarkSuit"): "Varia Suit",
    (RandovaniaGame.METROID_PRIME_ECHOES, "LightSuit"): "Varia Suit",
    (RandovaniaGame.METROID_PRIME_ECHOES, "MorphBall INCOMPLETE"): "Morph Ball",
    (RandovaniaGame.METROID_PRIME_ECHOES, "MorphBallBomb"): "Morph Ball Bomb",
    (RandovaniaGame.METROID_PRIME_ECHOES, "BoostBall"): "Boost Ball",
    (RandovaniaGame.METROID_PRIME_ECHOES, "SpiderBall"): "Spider Ball",
    (RandovaniaGame.METROID_PRIME_ECHOES, "PowerBomb"): "Power Bomb",
    (RandovaniaGame.METROID_PRIME_ECHOES, "PowerBombExpansion"): "Power Bomb Expansion",
    (RandovaniaGame.METROID_PRIME_ECHOES, "MissileExpansion"): "Missile",
    (RandovaniaGame.METROID_PRIME_ECHOES, "MissileExpansionPrime1"): "Missile",
    (RandovaniaGame.METROID_PRIME_ECHOES, "MissileLauncher"): "Missile",
    (RandovaniaGame.METROID_PRIME_ECHOES, "GrappleBeam"): "Grapple Beam",
    (RandovaniaGame.METROID_PRIME_ECHOES, "SpaceJumpBoots"): "Space Jump Boots",
    (RandovaniaGame.METROID_PRIME_ECHOES, "EnergyTank"): "Energy Tank",
}

# The following locations have cutscenes that weren't removed
_LOCATIONS_WITH_MODAL_ALERT = {
    63,  # Artifact Temple
    23,  # Watery Hall (Charge Beam)
    50,  # Research Core
}

# Show a popup on collection if two or more is for another player.
# The location to the right is considered for the count, but it can't show a popup.
_LOCATIONS_GROUPED_TOGETHER = [
    ({0, 1, 2, 3}, None),  # Main Plaza
    ({5, 6, 7}, None),  # Ruined Shrine (all 3)
    ({27, 28}, None),  # Burn dome
    ({94}, 97),  # Warrior shrine -> Fiery Shores Tunnel
    ({23, 24}, None),  # Watery Hall
    ({25, 26}, None),  # Dynamo (the one in Chozo)
    ({55}, 54),  # Gravity Chamber: Upper -> Lower
    ({19, 17}, None),  # Hive Totem + Transport Access North
    ({12}, 11),  # Magma Pool -> Training Chamber Access
    ({59}, 58),  # Alcove -> Landing Site
    ({62, 65}, None),  # Root Cave + Arbor Chamber
    ({40}, 39),  # Ice Ruins East (spider -> ice)
    ({15, 16}, None),  # Ruined Gallery
    ({18}, 22),  # Gathering Hall -> Watery Hall Access
    ({52, 53}, None),  # Research Lab Aether
]


def prime1_pickup_details_to_patcher(detail: pickup_exporter.ExportedPickupDetails,
                                     modal_hud_override: bool,
                                     rng: Random) -> dict:
    if detail.model.game == RandovaniaGame.METROID_PRIME:
        model_name = detail.model.name
    else:
        model_name = _MODEL_MAPPING.get((detail.model.game, detail.model.name), "Nothing")

    scan_text = detail.scan_text
    hud_text = detail.hud_text[0]
    pickup_type = "Nothing"
    count = 0

    for resource, quantity in detail.conditional_resources[0].resources:
        if resource.index >= 1000:
            continue
        pickup_type = resource.long_name
        count = quantity
        break

    if (model_name == "Missile" and not detail.other_player
            and "Missile Expansion" in hud_text
            and rng.randint(0, _EASTER_EGG_SHINY_MISSILE) == 0):
        model_name = "Shiny Missile"
        hud_text = hud_text.replace("Missile Expansion", "Shiny Missile Expansion")
        scan_text = scan_text.replace("Missile Expansion", "Shiny Missile Expansion")

    result = {
        "type": pickup_type,
        "model": model_name,
        "scanText": scan_text,
        "hudmemoText": hud_text,
        "currIncrease": count,
        "maxIncrease": count,
        "respawn": False
    }
    if modal_hud_override:
        result["modalHudmemo"] = True

    return result


def _create_locations_with_modal_hud_memo(pickups: List[pickup_exporter.ExportedPickupDetails]) -> typing.Set[int]:
    result = set()

    for index in _LOCATIONS_WITH_MODAL_ALERT:
        if pickups[index].other_player:
            result.add(index)

    for indices, extra in _LOCATIONS_GROUPED_TOGETHER:
        num_other = sum(pickups[i].other_player for i in indices)
        if extra is not None:
            num_other += pickups[extra].other_player

        if num_other > 1:
            for index in indices:
                if pickups[index].other_player:
                    result.add(index)

    return result


def _starting_items_value_for(resource_database: ResourceDatabase,
                              starting_items: CurrentResources, index: int) -> Union[bool, int]:
    item = resource_database.get_item(index)
    value = starting_items.get(item, 0)
    if item.max_capacity > 1:
        return value
    else:
        return value > 0


def _name_for_location(world_list: WorldList, location: AreaLocation) -> str:
    if location in prime1_elevators.RANDOM_PRIME_CUSTOM_NAMES:
        return prime1_elevators.RANDOM_PRIME_CUSTOM_NAMES[location]
    else:
        return world_list.area_name(world_list.area_by_area_location(location), separator=":")


class RandomprimePatcher(Patcher):
    _busy: bool = False

    @property
    def is_busy(self) -> bool:
        """
        Checks if the patcher is busy right now
        """
        return self._busy

    @property
    def export_can_be_aborted(self) -> bool:
        """
        Checks if patch_game can be aborted
        """
        return False

    @property
    def uses_input_file_directly(self) -> bool:
        """
        Does this patcher uses the input file directly?
        """
        return True

    def has_internal_copy(self, game_files_path: Path) -> bool:
        return False

    def delete_internal_copy(self, game_files_path: Path):
        pass

    def default_output_file(self, seed_hash: str) -> str:
        return "Prime Randomizer - {}".format(seed_hash)

    @property
    def valid_input_file_types(self) -> List[str]:
        return ["iso"]

    @property
    def valid_output_file_types(self) -> List[str]:
        return ["iso", "ciso", "gcz"]

    def create_patch_data(self, description: LayoutDescription, players_config: PlayersConfiguration,
                          cosmetic_patches: PrimeCosmeticPatches):
        patches = description.all_patches[players_config.player_index]
        db = default_database.game_description_for(RandovaniaGame.METROID_PRIME)
        preset = description.permalink.get_preset(players_config.player_index)
        configuration = typing.cast(PrimeConfiguration, preset.configuration)
        rng = Random(description.permalink.seed_number)

        area_namers = {
            index: hint_lib.AreaNamer(default_database.game_description_for(player_preset.game).world_list)
            for index, player_preset in description.permalink.presets.items()
        }

        scan_visor = db.resource_database.get_item_by_name("Scan Visor")
        useless_target = PickupTarget(pickup_creator.create_nothing_pickup(db.resource_database),
                                      players_config.player_index)

        pickup_list = pickup_exporter.export_all_indices(
            patches,
            useless_target,
            db.world_list,
            rng,
            configuration.pickup_model_style,
            configuration.pickup_model_data_source,
            exporter=pickup_exporter.create_pickup_exporter(db, pickup_exporter.GenericAcquiredMemo(), players_config),
            visual_etm=pickup_creator.create_visual_etm(),
        )
        modal_hud_override = _create_locations_with_modal_hud_memo(pickup_list)

        world_data = {}
        for world in db.world_list.worlds:
            if world.name == "End of Game":
                continue

            world_data[world.name] = {
                "transports": {},
                "rooms": {}
            }
            for area in world.areas:
                pickup_indices = sorted(node.pickup_index for node in area.nodes if isinstance(node, PickupNode))
                if pickup_indices:
                    world_data[world.name]["rooms"][area.name] = {
                        "pickups": [
                            prime1_pickup_details_to_patcher(pickup_list[index.index],
                                                             index.index in modal_hud_override,
                                                             rng)
                            for index in pickup_indices
                        ],
                    }

                for node in area.nodes:
                    if not isinstance(node, TeleporterNode) or not node.editable:
                        continue

                    target = _name_for_location(db.world_list, patches.elevator_connection[node.teleporter])
                    source_name = prime1_elevators.RANDOM_PRIME_CUSTOM_NAMES[node.teleporter.area_location]
                    world_data[world.name]["transports"][source_name] = target

        starting_memo = None
        extra_starting = item_names.additional_starting_items(configuration, db.resource_database,
                                                              patches.starting_items)
        if extra_starting:
            starting_memo = ", ".join(extra_starting)

        if cosmetic_patches.open_map and configuration.elevators.is_vanilla:
            map_default_state = "visible"
        else:
            map_default_state = "default"

        credits_string = credits_spoiler.prime_trilogy_credits(
            configuration.major_items_configuration,
            description.all_patches,
            players_config,
            area_namers,
            "&push;&font=C29C51F1;&main-color=#89D6FF;Major Item Locations&pop;",
            "&push;&font=C29C51F1;&main-color=#33ffd6;{}&pop;",
        )

        artifacts = [
            db.resource_database.get_item(index)
            for index in prime_items.ARTIFACT_ITEMS
        ]
        resulting_hints = guaranteed_item_hint.create_guaranteed_hints_for_resources(
            description.all_patches, players_config, area_namers, False,
            [db.resource_database.get_item(index) for index in prime_items.ARTIFACT_ITEMS],
            hint_lib.TextColor.PRIME1_ITEM,
            hint_lib.TextColor.PRIME1_LOCATION,
        )

        # Tweaks
        ctwk_config = {}
        if configuration.small_samus:
            ctwk_config["playerSize"] = 0.3
            ctwk_config["morphBallSize"] = 0.3
            ctwk_config["easyLavaEscape"] = True

        return {
            "seed": description.permalink.seed_number,
            "preferences": {
                "qolGameBreaking": configuration.qol_game_breaking,
                "qolCosmetic": cosmetic_patches.qol_cosmetic,
                "qolPickupScans": configuration.qol_pickup_scans,
                "qolCutscenes": configuration.qol_cutscenes.value,

                "mapDefaultState": map_default_state,
                "artifactHintBehavior": None,
                "automaticCrashScreen": True,
                "trilogyDiscPath": None,
                "quickplay": False,
                "quiet": False,
            },
            "gameConfig": {
                "startingRoom": _name_for_location(db.world_list, patches.starting_location),
                "startingMemo": starting_memo,
                "warpToStart": configuration.warp_to_start,

                "nonvariaHeatDamage": configuration.heat_protection_only_varia,
                "staggeredSuitDamage": configuration.progressive_damage_reduction,
                "heatDamagePerSec": configuration.heat_damage,
                "autoEnabledElevators": patches.starting_items.get(scan_visor, 0) == 0,
                "multiworldDolPatches": True,

                "startingItems": {
                    name: _starting_items_value_for(db.resource_database, patches.starting_items, index)
                    for name, index in _STARTING_ITEM_NAME_TO_INDEX.items()
                },

                "etankCapacity": configuration.energy_per_tank,
                "itemMaxCapacity": {
                    "Unknown Item 1": db.resource_database.multiworld_magic_item.max_capacity,
                },

                "mainPlazaDoor": configuration.main_plaza_door,
                "backwardsFrigate": configuration.backwards_frigate,
                "backwardsLabs": configuration.backwards_labs,
                "backwardsUpperMines": configuration.backwards_upper_mines,
                "backwardsLowerMines": configuration.backwards_lower_mines,
                "phazonEliteWithoutDynamo": configuration.phazon_elite_without_dynamo,

                "gameBanner": {
                    "gameName": "Metroid Prime: Randomizer",
                    "gameNameFull": "Metroid Prime: Randomizer - {}".format(description.shareable_hash),
                    "description": "Seed Hash: {}".format(description.shareable_word_hash),
                },
                "mainMenuMessage": "Randovania v{}\n{}".format(randovania.VERSION, description.shareable_word_hash),

                "creditsString": credits_string,
                "artifactHints": {
                    artifact.long_name: text
                    for artifact, text in resulting_hints.items()
                },
                "artifactTempleLayerOverrides": {
                    artifact.long_name: patches.starting_items.get(artifact, 0) == 0
                    for artifact in artifacts
                },
            },
            "tweaks": ctwk_config,
            "levelData": world_data,
            "hasSpoiler": description.permalink.spoiler,
        }

    def patch_game(self, input_file: Optional[Path], output_file: Path, patch_data: dict,
                   internal_copies_path: Path, progress_update: ProgressUpdateCallable):
        if input_file is None:
            raise ValueError("Missing input file")

        symbols = py_randomprime.symbols_for_file(input_file)

        new_config = copy.copy(patch_data)
        has_spoiler = new_config.pop("hasSpoiler")
        new_config["inputIso"] = os.fspath(input_file)
        new_config["outputIso"] = os.fspath(output_file)
        new_config["gameConfig"]["updateHintStateReplacement"] = list(
            assembler.assemble_instructions(
                symbols["UpdateHintState__13CStateManagerFf"],
                [
                    *all_prime_dol_patches.remote_execution_patch_start(),
                    *all_prime_dol_patches.remote_execution_patch_end(),
                ],
                symbols=symbols)
        )

        patch_as_str = json.dumps(new_config, indent=4, separators=(',', ': '))
        if has_spoiler:
            output_file.with_name(f"{output_file.stem}-patcher.json").write_text(patch_as_str)

        os.environ["RUST_BACKTRACE"] = "1"

        try:
            py_randomprime.patch_iso_raw(
                patch_as_str,
                py_randomprime.ProgressNotifier(lambda percent, msg: progress_update(msg, percent)),
            )
        except BaseException as e:
            if isinstance(e, Exception):
                raise
            else:
                raise RuntimeError(f"randomprime panic: {e}") from e



import struct

import pytest
from PySide2.QtCore import Qt
from mock import patch, MagicMock, AsyncMock

from randovania.game_connection.connection_base import InventoryItem
from randovania.game_connection.connector.prime_remote_connector import PrimeRemoteConnector, DolRemotePatch
from randovania.game_connection.executor.memory_operation import MemoryOperation
from randovania.games.prime import all_prime_dol_patches
from randovania.gui.debug_backend_window import DebugExecutorWindow


def _echoes_powerup_offset(item_index: int) -> int:
    powerups_offset = 0x58
    vector_data_offset = 0x4
    powerup_size = 0xc
    return (powerups_offset + vector_data_offset) + (item_index * powerup_size)


@pytest.fixture(name="executor")
def debug_executor_window(skip_qtbot):
    return DebugExecutorWindow()


@pytest.mark.asyncio
async def test_display_message(executor: DebugExecutorWindow):
    # Setup
    await executor._ensure_initialized_game_memory()
    connector = PrimeRemoteConnector(executor._used_version)
    message = "Foo Bar"

    # Run
    await connector.execute_remote_patches(
        executor,
        [connector._dol_patch_for_hud_message(message)],
    )
    executor._handle_remote_execution()

    # Assert
    assert executor.messages_list.findItems(message, Qt.MatchFlag.MatchExactly)


@pytest.mark.asyncio
async def test_update_inventory_label(executor: DebugExecutorWindow, echoes_resource_database):
    # Setup
    await executor._ensure_initialized_game_memory()
    connector = PrimeRemoteConnector(executor._used_version)

    await connector.execute_remote_patches(
        executor,
        [
            DolRemotePatch([], all_prime_dol_patches.increment_item_capacity_patch(
                executor._used_version.powerup_functions,
                echoes_resource_database.game_enum,
                echoes_resource_database.multiworld_magic_item.index,
                delta=2,
            )),
            DolRemotePatch([], all_prime_dol_patches.adjust_item_amount_and_capacity_patch(
                executor._used_version.powerup_functions,
                echoes_resource_database.game_enum,
                echoes_resource_database.energy_tank.index,
                delta=4,
            )),
        ],
    )
    executor._handle_remote_execution()

    # Run
    await executor._update_inventory_label()

    # Assert
    assert "Energy Tank x 4/4" in executor.inventory_label.text()
    assert "Multiworld Magic Identifier x 0/2" in executor.inventory_label.text()


@pytest.mark.asyncio
@patch("randovania.gui.lib.common_qt_lib.get_network_client", autospec=True)
async def test_setup_locations_combo(mock_get_network_client: MagicMock,
                                     executor: DebugExecutorWindow):
    # Setup
    patcher_data = {
        "pickups": [
            {"pickup_index": i, "hud_text": f"Item {i}"}
            for i in range(110)
        ]
    }
    mock_get_network_client.return_value.session_admin_player = AsyncMock(return_value=patcher_data)
    executor._used_version = executor._used_version

    # Run
    await executor._setup_locations_combo()

    # Assert
    mock_get_network_client.assert_called_once_with()
    assert executor.collect_location_combo.count() > 100


def test_clear(executor: DebugExecutorWindow):
    executor.clear()
    assert executor.messages_list.count() == 0
    assert executor.inventory_label.text() == ""



import dataclasses
import datetime
import json
from typing import List, Dict, Optional, Tuple

from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.binary_data import convert_to_raw_python
from randovania.games.game import RandovaniaGame
from randovania.layout.preset_migration import VersionedPreset
from randovania.network_common.binary_formats import BinaryGameSessionEntry
from randovania.network_common.session_state import GameSessionState


@dataclasses.dataclass(frozen=True)
class GameSessionListEntry:
    id: int
    name: str
    has_password: bool
    state: GameSessionState
    num_players: int
    creator: str
    creation_date: datetime.datetime

    def __post_init__(self):
        tzinfo = self.creation_date.tzinfo
        assert tzinfo is not None and tzinfo.utcoffset(self.creation_date) is not None

    @classmethod
    def from_json(cls, data: dict) -> "GameSessionListEntry":
        data["state"] = GameSessionState(data["state"])
        data["creation_date"] = datetime.datetime.fromisoformat(data["creation_date"])
        return GameSessionListEntry(**data)


@dataclasses.dataclass(frozen=True)
class PlayerSessionEntry:
    id: int
    name: str
    row: Optional[int]
    admin: bool
    connection_state: str

    @classmethod
    def from_json(cls, data) -> "PlayerSessionEntry":
        return PlayerSessionEntry(
            id=data["id"],
            name=data["name"],
            row=data["row"],
            admin=data["admin"],
            connection_state=data["connection_state"],
        )

    @property
    def is_observer(self):
        return self.row is None


@dataclasses.dataclass(frozen=True)
class GameSessionAction:
    provider: str
    provider_row: int
    receiver: str
    pickup: str
    location: PickupIndex
    time: datetime.datetime

    @classmethod
    def from_json(cls, data) -> "GameSessionAction":
        return GameSessionAction(
            provider=data["provider"],
            provider_row=data["provider_row"],
            receiver=data["receiver"],
            pickup=data["pickup"],
            location=PickupIndex(data["location"]),
            time=datetime.datetime.fromisoformat(data["time"]),
        )


@dataclasses.dataclass(frozen=True)
class GameDetails(JsonDataclass):
    seed_hash: str
    word_hash: str
    spoiler: bool
    permalink: str


@dataclasses.dataclass(frozen=True)
class GameSessionEntry:
    id: int
    name: str
    presets: List[VersionedPreset]
    players: Dict[int, PlayerSessionEntry]
    game_details: Optional[GameDetails]
    state: GameSessionState
    generation_in_progress: Optional[int]
    allowed_games: List[RandovaniaGame]

    @property
    def num_admins(self) -> int:
        return sum(1 for player in self.players.values() if player.admin)

    @property
    def num_rows(self) -> int:
        return len(self.presets)

    @classmethod
    def from_json(cls, data) -> "GameSessionEntry":
        data = convert_to_raw_python(BinaryGameSessionEntry.parse(data))

        player_entries = [
            PlayerSessionEntry.from_json(player_json)
            for player_json in data["players"]
        ]
        return GameSessionEntry(
            id=data["id"],
            name=data["name"],
            presets=[VersionedPreset(json.loads(preset_json)) for preset_json in data["presets"]],
            players={
                player_entry.id: player_entry
                for player_entry in player_entries
            },
            game_details=GameDetails.from_json(data["game_details"]) if data["game_details"] is not None else None,
            state=GameSessionState(data["state"]),
            generation_in_progress=data["generation_in_progress"],
            allowed_games=[RandovaniaGame(game) for game in data["allowed_games"]],
        )


@dataclasses.dataclass(frozen=True)
class GameSessionActions:
    actions: Tuple[GameSessionAction, ...]


@dataclasses.dataclass(frozen=True)
class GameSessionPickups:
    game: RandovaniaGame
    pickups: Tuple[Tuple[str, PickupEntry], ...]


@dataclasses.dataclass(frozen=True)
class GameSessionAuditEntry:
    user: str
    message: str
    time: datetime.datetime

    @classmethod
    def from_json(cls, data) -> "GameSessionAuditEntry":
        return GameSessionAuditEntry(
            user=data["user"],
            message=data["message"],
            time=datetime.datetime.fromisoformat(data["time"]),
        )


@dataclasses.dataclass(frozen=True)
class GameSessionAuditLog:
    entries: Tuple[GameSessionAuditEntry, ...]


@dataclasses.dataclass(frozen=True)
class User:
    id: int
    name: str

    @classmethod
    def from_json(cls, data) -> "User":
        return cls(
            id=data["id"],
            name=data["name"],
        )

    @property
    def as_json(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
        }



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, List

from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.lib.status_update_lib import ProgressUpdateCallable
from randovania.layout.layout_description import LayoutDescription


class Patcher(ABC):
    @property
    @abstractmethod
    def is_busy(self) -> bool:
        """
        Checks if the patcher is busy right now
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def export_can_be_aborted(self) -> bool:
        """
        Checks if patch_game can be aborted
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def uses_input_file_directly(self) -> bool:
        """
        Does this patcher uses the input file directly?
        """
        raise NotImplementedError()

    def has_internal_copy(self, game_files_path: Path) -> bool:
        """
        Checks if the internal storage has an usable copy of the game
        """
        raise NotImplementedError()

    def delete_internal_copy(self, game_files_path: Path):
        """
        Deletes any copy of the game in the internal storage.
        """
        raise NotImplementedError()

    def default_output_file(self, seed_hash: str) -> str:
        """
        Provides a output file name with the given seed hash.
        :param seed_hash:
        :return:
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def valid_input_file_types(self) -> List[str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def valid_output_file_types(self) -> List[str]:
        raise NotImplementedError()

    def create_patch_data(self, description: LayoutDescription, players_config: PlayersConfiguration,
                          cosmetic_patches: BaseCosmeticPatches) -> dict:
        """
        Creates a JSON serializable dict that can be used to patch the game.
        Intended to be ran on the server for multiworld.
        :return:
        """
        raise NotImplementedError()

    def patch_game(self, input_file: Optional[Path], output_file: Path, patch_data: dict,
                   internal_copies_path: Path, progress_update: ProgressUpdateCallable):
        """
        Invokes the necessary tools to patch the game.
        :param input_file: Vanilla copy of the game. Required if uses_input_file_directly or has_internal_copy is False.
        :param output_file: Where a modified copy of the game is placed.
        :param patch_data: Data created by create_patch_data.
        :param internal_copies_path: Path to where all internal copies are stored.
        :param progress_update: Pushes updates as slow operations are done.
        :return: None
        """
        raise NotImplementedError()



import logging
import multiprocessing
import sys


logging.basicConfig(level=logging.WARNING)


def main():
    multiprocessing.freeze_support()

    import randovania
    randovania.setup_logging('INFO', None)

    logging.info("Starting Randovania...")

    from randovania import cli
    cli.run_cli(sys.argv)


if __name__ == "__main__":
    main()



import dataclasses
from randovania.interface_common.preset_editor import PresetEditor
from PySide2 import QtWidgets

from randovania.layout.preset import Preset


class PresetTab(QtWidgets.QMainWindow):
    def __init__(self, editor: PresetEditor) -> None:
        super().__init__()
        self._editor = editor

    @property
    def tab_title(self):
        return self.windowTitle()

    @property
    def uses_patches_tab(self) -> bool:
        raise NotImplementedError()

    def on_preset_changed(self, preset: Preset):
        raise NotImplementedError()

    # Persistence helpers
    def _persist_enum(self, combo: QtWidgets.QComboBox, attribute_name: str):
        def persist(index: int):
            with self.editor as options:
                options.set_configuration_field(attribute_name, combo.itemData(index))
        return persist

    def _persist_bool_layout_field(self, field_name: str):
        def bound(value: int):
            with self.editor as options:
                options.set_configuration_field(field_name, bool(value))
        return bound

    def _persist_bool_major_configuration_field(self, field_name: str):
        def bound(value: int):
            with self.editor as options:
                kwargs = {field_name: bool(value)}
                options.major_items_configuration = dataclasses.replace(options.major_items_configuration, **kwargs)
        return bound

    def _persist_option_then_notify(self, attribute_name: str):
        def persist(value: int):
            with self._editor as options:
                setattr(options, attribute_name, bool(value))

        return persist

    def _persist_float(self, attribute_name: str):
        def persist(value: float):
            with self._editor as options:
                options.set_configuration_field(attribute_name, value)

        return persist



import dataclasses
import uuid
import pytest
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.node import PickupNode
from randovania.gui.preset_settings.location_pool_row_widget import LocationPoolRowWidget
from randovania.gui.preset_settings.location_pool_tab import PresetLocationPool
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset

@pytest.fixture()
def pickup_node():
    return PickupNode(
        pickup_index=PickupIndex(1), 
        major_location=True,
        name="Pickup (Ultra Beam)",
        heal=False,
        location=None,
        index=0
    )


def test_location_pool_row_initial_state(pickup_node, skip_qtbot):
    # Setup & Run
    widget = LocationPoolRowWidget(pickup_node, "Fancy name for a pickup")
    skip_qtbot.addWidget(widget)

    # Assert
    assert widget.label_location_name.text() == "Fancy name for a pickup"


def test_location_pool_row_actions(skip_qtbot):
    # Setup
    widget = LocationPoolRowWidget(pickup_node, "Fancy name for a pickup")
    skip_qtbot.addWidget(widget)

    signal_received = False
    def edit_closure():
        nonlocal signal_received
        signal_received = True
    widget.changed.connect(edit_closure)

    # Run & Assert
    assert not signal_received
    widget.set_can_have_progression(True)
    assert signal_received
    assert widget.radio_shuffled.isChecked()



def test_location_pool_row_disabled_on_major_minor_split(customized_preset, echoes_game_description, skip_qtbot):
    # Setup
    preset_editor = PresetEditor(customized_preset)

    location_pool_tab = PresetLocationPool(preset_editor, echoes_game_description)

    # Get the first major location in the list, and the first non-major one
    # Then, put it in a state which will have to be changed in the case where
    # major/minor split is enabled.
    first_major: LocationPoolRowWidget = None
    first_non_major: LocationPoolRowWidget = None
    for row_widget in location_pool_tab._row_widget_for_node.values():
        if first_major is None and row_widget.node.major_location:
            first_major = row_widget
        elif first_non_major is None and not row_widget.node.major_location:
            first_non_major = row_widget
        if first_major is not None and first_non_major is not None:
            break
    first_major.radio_shuffled_no_progression.setChecked(True)

    # Run & Assert
    assert first_major.isEnabled()
    assert first_non_major.isEnabled()
    assert not first_major.radio_shuffled.isChecked()

    location_pool_tab.check_major_minor.setChecked(True)

    assert first_major.isEnabled()
    assert not first_non_major.isEnabled()
    assert first_major.radio_shuffled.isChecked()
